import pytest

from emmett_core.html import MetaHtmlTag, TagStack, TreeHtmlTag, cat


@pytest.fixture(scope="module")
def tag():
    stack = TagStack()
    return MetaHtmlTag(stack)


@pytest.fixture(scope="module")
def ttag():
    class MetaTreeHtmlTag(MetaHtmlTag):
        _tag_cls = TreeHtmlTag

    stack = TagStack()
    return MetaTreeHtmlTag(stack)


def test_tag_self_closed(tag):
    br = tag.br()
    assert str(br) == "<br />"


def test_tag_non_closed(tag):
    p = tag.p()
    assert str(p) == "<p></p>"


def test_tag_components(tag):
    t = tag.div(tag.p(), tag.p())
    assert str(t) == "<div><p></p><p></p></div>"


def test_tag_attributes(tag):
    d = tag.div(_class="test", _id="test", _test="test")
    assert str(d) == '<div class="test" id="test" test="test"></div>'


def test_tag_attributes_dict(tag):
    d = tag.div(_class="test", _hx={"foo": "bar"})
    assert str(d) == '<div class="test" hx-foo="bar"></div>'


def test_tag_attributes_data(tag):
    d = tag.div(_data={"foo": "bar"})
    assert str(d) == '<div data-foo="bar"></div>'


def test_cat(tag):
    t = cat(tag.p(), tag.p())
    assert str(t) == "<p></p><p></p>"


def test_tag_stack(tag):
    t = tag.div(_class="l0")
    with t:
        with tag.div(_class="l1"):
            tag.p("t1")
            with tag.p() as p:
                p.append("t2")
        with tag.div(_class="l1"):
            tag.p("t3")
    assert str(t) == '<div class="l0"><div class="l1"><p>t1</p><p>t2</p></div><div class="l1"><p>t3</p></div></div>'


def test_tag_tree_stack(ttag):
    t = ttag.div(_class="l0")
    with t:
        with ttag.div(_class="l1") as d1:
            assert d1.parent is t
            p1 = ttag.p("t1")
            assert p1.parent is d1
            with ttag.p() as p2:
                assert p2.parent is d1
                p2.append("t2")
        with ttag.div(_class="l1") as d2:
            assert d2.parent is t
            p3 = ttag.p("t3")
            assert p3.parent is d2
    assert str(t) == '<div class="l0"><div class="l1"><p>t1</p><p>t2</p></div><div class="l1"><p>t3</p></div></div>'



import asyncio
from typing import AsyncGenerator

from ...http.response import HTTPBytesResponse, HTTPResponse
from ...http.wrappers.response import ResponseStream as _ResponseStream


class BodyWrapper:
    __slots__ = ["proto", "timeout"]

    def __init__(self, proto, timeout):
        self.proto = proto
        self.timeout = timeout

    def __await__(self):
        if self.timeout:
            return self._await_with_timeout().__await__()
        return self.proto().__await__()

    async def _await_with_timeout(self):
        try:
            rv = await asyncio.wait_for(self.proto(), timeout=self.timeout)
        except asyncio.TimeoutError:
            raise HTTPBytesResponse(408, b"Request timeout")
        return rv

    async def __aiter__(self) -> AsyncGenerator[bytes, None]:
        async for chunk in self.proto:
            yield chunk


class ResponseStream(_ResponseStream):
    __slots__ = []

    def __call__(self):
        ctl_event = asyncio.Event()
        task_stream = asyncio.create_task(self._handle_stream(ctl_event))
        task_transport = asyncio.create_task(self._handle_conn(self._proto, task_stream, ctl_event))
        return self._control_flow(ctl_event, task_transport)

    async def _handle_stream(self, ctl_event):
        for method in self.response._flow_stream:
            method()
        transport = self._proto.response_stream(self.response.status, list(HTTPResponse.rsgi_headers(self)))
        async for item in self._target:
            await self.send(transport, self._item_wrapper(item))
        ctl_event.set()

    async def _handle_conn(self, protocol, stream_task, ctl_event):
        if ctl_event.is_set():
            return
        await protocol.client_disconnect()
        if ctl_event.is_set():
            return
        stream_task.cancel()
        ctl_event.set()

    async def _control_flow(self, event, transport_task):
        await event.wait()
        transport_task.cancel()
        return noop_response

    def send(self, transport, data):
        if isinstance(data, str):
            return transport.send_str(data)
        return transport.send_bytes(data)


class NoopResponse:
    def rsgi(self, protocol):
        return


class WSTransport:
    __slots__ = ["protocol", "transport", "accepted", "interrupted", "input", "status", "noop"]

    def __init__(self, protocol) -> None:
        self.protocol = protocol
        self.transport = None
        self.accepted = asyncio.Event()
        self.input = asyncio.Queue()
        self.interrupted = False
        self.status = 200
        self.noop = asyncio.Event()

    async def init(self):
        self.transport = await self.protocol.accept()
        self.accepted.set()

    @property
    def receive(self):
        return self.input.get


noop_response = NoopResponse()



from typing import Any, Awaitable, Callable, Dict, Tuple


Scope = Dict[str, Any]
Receive = Callable[[], Awaitable[Dict[str, Any]]]
Send = Callable[[Dict[str, Any]], Awaitable[None]]
Event = Dict[str, Any]
EventHandler = Callable[[Any, Scope, Receive, Send, Event], Awaitable[Any]]
EventLooper = Callable[..., Awaitable[Tuple[EventHandler, Event]]]



from ._imports import granian


def run(
    interface,
    app,
    host="127.0.0.1",
    port=8000,
    loop="auto",
    task_impl="asyncio",
    log_level=None,
    log_access=False,
    workers=1,
    runtime_threads=1,
    runtime_blocking_threads=None,
    runtime_mode="st",
    backlog=1024,
    backpressure=None,
    http="auto",
    http_read_timeout=10_000,
    enable_websockets=True,
    ssl_certfile=None,
    ssl_keyfile=None,
    reload=False,
    **kwargs,
):
    if granian is None:
        raise RuntimeError("granian dependency not installed")

    http1_settings = granian.http.HTTP1Settings(header_read_timeout=http_read_timeout)
    http2_settings = granian.http.HTTP2Settings(keep_alive_interval=http_read_timeout)

    app_path = ":".join([app[0], app[1] or "app"])
    server = granian.Granian(
        app_path,
        address=host,
        port=port,
        interface=interface,
        workers=workers,
        runtime_threads=runtime_threads,
        runtime_blocking_threads=runtime_blocking_threads,
        runtime_mode=runtime_mode,
        loop=loop,
        task_impl=task_impl,
        http=http,
        websockets=enable_websockets,
        backlog=backlog,
        backpressure=backpressure,
        http1_settings=http1_settings,
        http2_settings=http2_settings,
        log_level=log_level,
        log_access=log_access,
        ssl_cert=ssl_certfile,
        ssl_key=ssl_keyfile,
        reload=reload,
    )
    server.serve()



import logging
import os
from logging import Formatter, StreamHandler
from logging.handlers import RotatingFileHandler
from threading import Lock

from .datastructures import sdict


_logger_lock = Lock()

LOG_LEVELS = {
    "debug": logging.DEBUG,
    "info": logging.INFO,
    "warning": logging.WARNING,
    "error": logging.ERROR,
    "critical": logging.CRITICAL,
}

_def_log_config = sdict(
    production=sdict(
        level="warning", format="[%(asctime)s] %(levelname)s in %(module)s: %(message)s", on_app_debug=False
    )
)
_debug_log_format = "> %(levelname)s in %(module)s [%(pathname)s:%(lineno)d]:\n" + "%(message)s"


def create_logger(app):
    Logger = logging.getLoggerClass()

    class DebugLogger(Logger):
        def getEffectiveLevel(self):
            if self.level == 0 and app.debug:
                return logging.DEBUG
            return Logger.getEffectiveLevel(self)

    class DebugHandlerSTD(StreamHandler):
        def emit(self, record):
            StreamHandler.emit(self, record) if app.debug else None

    class HandlerSTD(StreamHandler):
        def emit(self, record):
            StreamHandler.emit(self, record) if not app.debug else None

    class DebugHandlerRF(RotatingFileHandler):
        def emit(self, record):
            RotatingFileHandler.emit(self, record) if app.debug else None

    class HandlerRF(RotatingFileHandler):
        def emit(self, record):
            RotatingFileHandler.emit(self, record) if not app.debug else None

    #: init the console debug handler
    debug_handler = DebugHandlerSTD()
    debug_handler.setLevel(logging.DEBUG)
    debug_handler.setFormatter(Formatter(_debug_log_format))
    logger = logging.getLogger(app.logger_name)
    #: just in case that was not a new logger, get rid of all the handlers already attached to it
    del logger.handlers[:]
    logger.__class__ = DebugLogger
    logger.addHandler(debug_handler)
    #: load application logging config
    app_logs = app.config.logging
    if not app_logs:
        app_logs = _def_log_config
    for lname, lconf in app_logs.items():
        level = LOG_LEVELS.get(lconf.level or "warning", LOG_LEVELS.get("warning"))
        lformat = lconf.format or _def_log_config.production.format
        if filecfg := lconf.file:
            lfile = os.path.join(app.root_path, "logs", lname + ".log")
            max_size = filecfg.max_size or 5 * 1024 * 1024
            file_no = filecfg.no or 4
            handler_cls = DebugHandlerRF if lconf.on_app_debug else HandlerRF
            handler = handler_cls(lfile, maxBytes=max_size, backupCount=file_no)
        else:
            handler_cls = DebugHandlerSTD if lconf.on_app_debug else HandlerSTD
            handler = handler_cls()
        handler.setLevel(level)
        handler.setFormatter(Formatter(lformat))
        logger.addHandler(handler)
    return logger



import pytest

from emmett_core.app import App as _App
from emmett_core.ctx import Current, RequestContext
from emmett_core.http.wrappers.response import Response as _Response
from emmett_core.protocols.rsgi.test_client.scope import ScopeBuilder
from emmett_core.protocols.rsgi.wrappers import Request
from emmett_core.routing.router import HTTPRouter, WebsocketRouter


class Response(_Response):
    async def stream(self, target, item_wrapper=None):
        raise NotImplementedError


class App(_App):
    def _init_routers(self, url_prefix):
        pass

    def _init_handlers(self):
        pass

    def _register_with_ctx(self):
        pass

    def _init_with_test_env(self, current):
        self._router_http = HTTPRouter(self, current)
        self._router_ws = WebsocketRouter(self, current)
        current.app = self


@pytest.fixture(scope="function")
def current():
    return Current()


@pytest.fixture(scope="function")
def app(current):
    rv = App(__name__)
    rv._init_with_test_env(current)
    return rv


@pytest.fixture(scope="function")
def http_ctx(current, app):
    scope = ScopeBuilder(path="/").get_data()[0]
    token = current._init_(RequestContext(app, Request(scope, scope.path, None), Response(None)))
    yield current
    current._close_(token)



from __future__ import annotations

import os
from typing import Any, Callable

from ..ctx import Current
from ..pipeline import Pipe, RequestPipeline, WebsocketPipeline
from .cache import RouteCacheRule
from .routes import HTTPRoute, WebsocketRoute


ALLOWED_SCHEMES = {"http", "https"}


class RoutingRule:
    __slots__ = ["router"]

    def __init__(self, router, *args, **kwargs):
        self.router = router

    @property
    def app(self):
        return self.router.app

    def build_name(self, f):
        filename = os.path.realpath(f.__code__.co_filename)
        short = filename[1 + len(self.app.root_path) :].rsplit(".", 1)[0]
        if not short:
            short = filename.rsplit(".", 1)[0]
        if short == "__init__":
            short = self.app.root_path.rsplit("/", 1)[-1]
        #: allow only one naming level if name is not provided
        if len(short.split(os.sep)) > 1:
            short = short.split(os.sep)[-1]
        return ".".join(short.split(os.sep) + [f.__name__])

    def __call__(self, f: Callable[..., Any]) -> Callable[..., Any]:
        raise NotImplementedError


class HTTPRoutingRule(RoutingRule):
    __slots__ = [
        "cache_rule",
        "f",
        "head_builder",
        "hostname",
        "methods",
        "name",
        "output_type",
        "paths",
        "pipeline_flow_close",
        "pipeline_flow_open",
        "pipeline_flow_stream",
        "pipeline",
        "prefix",
        "response_builder",
        "schemes",
    ]
    current: Current
    route_cls = HTTPRoute

    def __init__(
        self,
        router,
        paths=None,
        name=None,
        pipeline=None,
        schemes=None,
        hostname=None,
        methods=None,
        prefix=None,
        cache=None,
        output="auto",
    ):
        super().__init__(router)
        self.name = name
        self.paths = paths
        if self.paths is None:
            self.paths = []
        if not isinstance(self.paths, (list, tuple)):
            self.paths = (self.paths,)
        self.schemes = schemes or ("http", "https")
        if not isinstance(self.schemes, (list, tuple)):
            self.schemes = (self.schemes,)
        if not set(self.schemes).issubset(ALLOWED_SCHEMES):
            raise SyntaxError(f"Invalid schemes specified. Allowed values are: {', '.join(ALLOWED_SCHEMES)}")
        self.methods = methods or ("get", "post", "head")
        if not isinstance(self.methods, (list, tuple)):
            self.methods = (self.methods,)
        self.hostname = hostname or self.app.config.hostname_default
        if prefix:
            if not prefix.startswith("/"):
                prefix = "/" + prefix
        self.prefix = prefix
        if output not in self.router._outputs:
            raise SyntaxError(f"Invalid output specified. Allowed values are: {', '.join(self.router._outputs.keys())}")
        self.output_type = output
        self.pipeline = self.router.pipeline + (pipeline or [])
        self.cache_rule = None
        if cache:
            if not isinstance(cache, RouteCacheRule):
                raise RuntimeError("route cache argument should be a valid caching rule")
            if any(key in self.methods for key in ["get", "head"]):
                self.cache_rule = cache
        # check pipes are indeed valid pipes
        if any(not isinstance(pipe, Pipe) for pipe in self.pipeline):
            raise RuntimeError("Invalid pipeline")

    def _make_builders(self, output_type):
        builder_cls = self.router._outputs[output_type]
        return builder_cls(self), self.router._outputs["empty"](self)

    def __call__(self, f: Callable[..., Any]) -> Callable[..., Any]:
        if not self.paths:
            self.paths.append("/" + f.__name__)
        if not self.name:
            self.name = self.build_name(f)
        if self.name.endswith("."):
            self.name = self.name + f.__name__
        pipeline_obj = RequestPipeline(self.pipeline)
        wrapped_f = pipeline_obj(f)
        self.pipeline_flow_open = pipeline_obj._flow_open()
        self.pipeline_flow_close = pipeline_obj._flow_close()
        self.pipeline_flow_stream = pipeline_obj._flow_stream()
        self.f = wrapped_f
        output_type = pipeline_obj._output_type() or self.output_type
        self.response_builder, self.head_builder = self._make_builders(output_type)
        for idx, path in enumerate(self.paths):
            self.router.add_route(self.__class__.route_cls(self, path, idx))
        return f


class WebsocketRoutingRule(RoutingRule):
    __slots__ = [
        "f",
        "hostname",
        "name",
        "paths",
        "pipeline_flow_close",
        "pipeline_flow_open",
        "pipeline_flow_receive",
        "pipeline_flow_send",
        "pipeline",
        "prefix",
        "schemes",
    ]
    route_cls = WebsocketRoute

    def __init__(self, router, paths=None, name=None, pipeline=None, schemes=None, hostname=None, prefix=None):
        super().__init__(router)
        self.name = name
        self.paths = paths
        if self.paths is None:
            self.paths = []
        if not isinstance(self.paths, (list, tuple)):
            self.paths = (self.paths,)
        self.schemes = schemes or ("http", "https")
        if not isinstance(self.schemes, (list, tuple)):
            self.schemes = (self.schemes,)
        if not set(self.schemes).issubset(ALLOWED_SCHEMES):
            raise SyntaxError(f"Invalid schemes specified. Allowed values are: {', '.join(ALLOWED_SCHEMES)}")
        self.hostname = hostname or self.app.config.hostname_default
        if prefix:
            if not prefix.startswith("/"):
                prefix = "/" + prefix
        self.prefix = prefix
        self.pipeline = self.router.pipeline + (pipeline or [])
        # check pipes are indeed valid pipes
        if any(not isinstance(pipe, Pipe) for pipe in self.pipeline):
            raise RuntimeError("Invalid pipeline")

    def __call__(self, f: Callable[..., Any]) -> Callable[..., Any]:
        if not self.paths:
            self.paths.append("/" + f.__name__)
        if not self.name:
            self.name = self.build_name(f)
        if self.name.endswith("."):
            self.name = self.name + f.__name__
        pipeline_obj = WebsocketPipeline(self.pipeline)
        wrapped_f = pipeline_obj(f)
        self.pipeline_flow_open = pipeline_obj._flow_open()
        self.pipeline_flow_close = pipeline_obj._flow_close()
        self.pipeline_flow_receive = pipeline_obj._flow_receive()
        self.pipeline_flow_send = pipeline_obj._flow_send()
        self.f = wrapped_f
        for idx, path in enumerate(self.paths):
            self.router.add_route(self.__class__.route_cls(self, path, idx))
        return f



from __future__ import annotations

import re
from abc import ABCMeta, abstractmethod
from http.cookies import SimpleCookie

from ..response import HTTPAsyncIterResponse, HTTPFileResponse, HTTPIOResponse, HTTPIterResponse, HTTPResponse
from . import EgressWrapper
from .helpers import ResponseHeaders


_SSE_NEWLINES_RE = re.compile(r"\r\n|\n")


class Response(EgressWrapper):
    __slots__ = ["_flow_stream", "status", "headers", "cookies"]

    def __init__(self, proto):
        super().__init__(proto)
        self.status = 200
        self.headers = ResponseHeaders({"content-type": "text/plain"})
        self.cookies = SimpleCookie()

    def _bind_flow(self, flow_stream):
        self._flow_stream = flow_stream

    @property
    def content_type(self) -> str:
        return self.headers["content-type"]

    @content_type.setter
    def content_type(self, value: str):
        self.headers["content-type"] = value

    def wrap_iter(self, obj) -> HTTPIterResponse:
        return HTTPIterResponse(obj, status_code=self.status, headers=self.headers, cookies=self.cookies)

    def wrap_aiter(self, obj) -> HTTPAsyncIterResponse:
        return HTTPAsyncIterResponse(obj, status_code=self.status, headers=self.headers, cookies=self.cookies)

    def wrap_file(self, path) -> HTTPFileResponse:
        return HTTPFileResponse(str(path), status_code=self.status, headers=self.headers, cookies=self.cookies)

    def wrap_io(self, obj, chunk_size: int = 4096) -> HTTPIOResponse:
        return HTTPIOResponse(
            obj, status_code=self.status, headers=self.headers, cookies=self.cookies, chunk_size=chunk_size
        )

    @abstractmethod
    async def stream(self, target, item_wrapper=None) -> HTTPResponse: ...


class ResponseStream(metaclass=ABCMeta):
    __slots__ = ["response", "_proto", "_target", "_item_wrapper"]

    def __init__(self, response: Response, target, item_wrapper=None):
        self.response = response
        self._proto = response._proto
        self._target = target
        self._item_wrapper = item_wrapper or (lambda v: v)

    @property
    def _headers(self):
        return self.response.headers

    @property
    def _cookies(self):
        return self.response.cookies

    def __await__(self):
        return self().__await__()

    @abstractmethod
    async def __call__(self): ...

    @abstractmethod
    async def send(self, data): ...


class ServerSentEvent:
    __slots__ = ["data", "event", "id", "retry", "comment"]

    def __init__(
        self,
        data,
        event=None,
        id=None,
        retry=-1,
        comment=None,
    ):
        self.data = data
        self.event = event
        self.id = id
        self.retry = retry
        self.comment = comment

    def encode(self, json_encoder) -> bytes:
        stack = bytearray()

        if self.id:
            stack.extend(b"id: " + _SSE_NEWLINES_RE.sub("", self.id).encode("utf8") + b"\r\n")
        if self.comment:
            for part in _SSE_NEWLINES_RE.split(self.comment):
                stack.extend(b": " + part.encode("utf8") + b"\r\n")
        if self.event:
            stack.extend(b"event: " + _SSE_NEWLINES_RE.sub("", self.event).encode("utf8") + b"\r\n")
        if self.data:
            data = self.data
            if not isinstance(data, (bytes, str)):
                data = json_encoder(data)
            if not isinstance(data, bytes):
                data = data.encode("utf8")
            stack.extend(b"data: " + data + b"\r\n")
        if self.retry > -1:
            stack.extend(b"retry: " + str(self.retry).encode() + b"\r\n")

        stack.extend(b"\r\n")
        return bytes(stack)



from typing import Any, Dict

from ..ctx import Current
from ..http.helpers import redirect
from ..http.wrappers.response import ServerSentEvent
from ..parsers import Parsers
from ..serializers import Serializers, _json_type
from .pipe import Pipe


class RequirePipe(Pipe):
    __slots__ = ["condition", "otherwise"]
    _current: Current

    def __init__(self, condition=None, otherwise=None):
        if condition is None or otherwise is None:
            raise SyntaxError("usage: @requires(condition, otherwise)")
        if not callable(otherwise) and not isinstance(otherwise, str):
            raise SyntaxError("'otherwise' param must be string or callable")
        self.condition = condition
        self.otherwise = otherwise

    async def pipe_request(self, next_pipe, **kwargs):
        flag = self.condition()
        if not flag:
            if self.otherwise is not None:
                if callable(self.otherwise):
                    return self.otherwise()
                redirect(self.__class__._current, self.otherwise)
            else:
                redirect(self.__class__._current, "/")
        return await next_pipe(**kwargs)

    async def pipe_ws(self, next_pipe, **kwargs):
        flag = self.condition()
        if not flag:
            return
        await next_pipe(**kwargs)


class JSONPipe(Pipe):
    __slots__ = ["decoder", "encoder"]
    _current: Current
    output = _json_type

    def __init__(self):
        self.decoder = Parsers.get_for("json")
        self.encoder = Serializers.get_for("json")

    async def pipe_request(self, next_pipe, **kwargs):
        self.__class__._current.response.headers._data["content-type"] = "application/json"
        return self.encoder(await next_pipe(**kwargs))

    def on_receive(self, data):
        return self.decoder(data)

    def on_send(self, data):
        return self.encoder(data)


class StreamPipe(Pipe):
    _current: Current

    def __init__(
        self, status: int = 200, headers: Dict[str, str] = {}, cookies: Dict[str, Any] = {}, item_wrapper: Any = None
    ):
        self.response_status = status
        self.response_headers = headers
        self.response_cookies = cookies
        self.item_wrapper = item_wrapper

    async def pipe_request(self, next_pipe, **kwargs):
        response = self.__class__._current.response
        response.status = self.response_status
        for key, val in self.response_headers.items():
            response.headers[key] = val
        for key, val in self.response_cookies.items():
            response.cookies[key] = val
        return await response.stream(next_pipe(**kwargs), item_wrapper=self.item_wrapper)


class SSEPipe(StreamPipe):
    def __init__(
        self,
        status: int = 200,
        headers: Dict[str, str] = {},
        cookies: Dict[str, Any] = {},
    ):
        self._json_encoder = Serializers.get_for("json")
        default_headers = {"content-type": "text/event-stream"}
        if not headers:
            default_headers["cache-control"] = "no-cache"
            default_headers["connection"] = "keep-alive"
        super().__init__(status, {**headers, **default_headers}, cookies, item_wrapper=self._stream_item_wrapper)

    def _stream_item_wrapper(self, item):
        if isinstance(item, dict):
            item = ServerSentEvent(**item)
        if isinstance(item, ServerSentEvent):
            return item.encode(self._json_encoder)
        if isinstance(item, bytes):
            return item
        return str(item).encode("utf8")



from datetime import datetime
from typing import Any, Dict, List, Optional, Union
from urllib.parse import parse_qs

from ...datastructures import sdict
from ...http.response import HTTPBytesResponse, HTTPResponse
from ...http.wrappers.helpers import regex_client
from ...http.wrappers.request import Request as _Request
from ...http.wrappers.response import Response as _Response
from ...http.wrappers.websocket import Websocket as _Websocket
from ...utils import cachedprop
from .helpers import BodyWrapper, ResponseStream


class RSGIIngressMixin:
    def __init__(self, scope, path, protocol):
        self._scope = scope
        self._proto = protocol
        self.scheme = scope.scheme
        self.path: str = path

    @property
    def headers(self):
        return self._scope.headers

    @cachedprop
    def host(self) -> str:
        if self._scope.http_version[0] == "1":
            return self.headers.get("host")
        return self._scope.authority

    @cachedprop
    def query_params(self) -> sdict[str, Union[str, List[str]]]:
        rv: sdict[str, Any] = sdict()
        for key, values in parse_qs(self._scope.query_string, keep_blank_values=True).items():
            if len(values) == 1:
                rv[key] = values[0]
                continue
            rv[key] = values
        return rv

    @cachedprop
    def client(self) -> str:
        g = regex_client.search(self.headers.get("x-forwarded-for", ""))
        client = (
            (g.group() or "").split(",")[0] if g else (self._scope.client.split(":")[0] if self._scope.client else None)
        )
        if client in (None, "", "unknown", "localhost"):
            client = "::1" if self.host.startswith("[") else "127.0.0.1"
        return client  # type: ignore


class Request(RSGIIngressMixin, _Request):
    __slots__ = ["_scope", "_proto"]

    def __init__(
        self,
        scope,
        path,
        protocol,
        max_content_length: Optional[int] = None,
        max_multipart_size: Optional[int] = None,
        body_timeout: Optional[int] = None,
    ):
        super().__init__(scope, path, protocol)
        self.max_content_length = max_content_length
        self.max_multipart_size = max_multipart_size
        self.body_timeout = body_timeout
        self._now = datetime.utcnow()
        self.method = scope.method

    @property
    def _multipart_headers(self):
        return dict(self.headers.items())

    @cachedprop
    def body(self) -> BodyWrapper:
        if self.max_content_length and self.content_length > self.max_content_length:
            raise HTTPBytesResponse(413, b"Request entity too large")
        return BodyWrapper(self._proto, self.body_timeout)

    async def push_promise(self, path: str):
        raise NotImplementedError("RSGI protocol doesn't support HTTP2 push.")


class Response(_Response):
    async def stream(self, target, item_wrapper=None) -> HTTPResponse:
        return await ResponseStream(self, target, item_wrapper=item_wrapper)


class Websocket(RSGIIngressMixin, _Websocket):
    __slots__ = ["_scope", "_proto"]

    def __init__(self, scope, path, protocol):
        super().__init__(scope, path, protocol)
        self._flow_receive = None
        self._flow_send = None
        self.receive = self._accept_and_receive
        self.send = self._accept_and_send

    async def accept(self, headers: Optional[Dict[str, str]] = None, subprotocol: Optional[str] = None):
        if self._proto.transport:
            return
        await self._proto.init()
        self.receive = self._wrapped_receive
        self.send = self._wrapped_send

    async def _wrapped_receive(self) -> Any:
        data = (await self._proto.receive()).data
        for method in self._flow_receive:
            data = method(data)
        return data

    async def _wrapped_send(self, data: Any):
        for method in self._flow_send:
            data = method(data)
        trx = self._proto.transport.send_str if isinstance(data, str) else self._proto.transport.send_bytes
        try:
            await trx(data)
        # except ProtocolClosed:
        except RuntimeError:
            if not self._proto.interrupted:
                raise
            await self._proto.noop.wait()



from __future__ import annotations

import hashlib
from typing import Any, Callable, Dict, List, Tuple


hashlib_sha1 = lambda s: hashlib.sha1(bytes(s, "utf8"))  # noqa: S324


class CacheHashMixin:
    def __init__(self):
        self.strategies = {}

    def add_strategy(self, key: str, method: Callable[..., Any] = lambda data: data):
        self.strategies[key] = method

    def _hash_component(self, key: str, data: Any) -> str:
        return "".join([key, "{", repr(data), "}"])

    def _build_hash(self, data: Dict[str, Any]) -> str:
        components = []
        for key, strategy in self.strategies.items():
            components.append(self._hash_component(key, strategy(data[key])))
        return hashlib_sha1(":".join(components)).hexdigest()

    def _build_ctx_key(self, **ctx) -> str:
        return self.key + ":" + self._build_hash(ctx)  # type: ignore

    @staticmethod
    def dict_strategy(data: Dict[str, Any]) -> List[Tuple[str, Any]]:
        return [(key, data[key]) for key in sorted(data)]



from functools import wraps

from ..http.response import HTTPResponse


def _wrap_flow_request_complete(pipe_method, on_success, on_failure, f):
    @wraps(f)
    async def flow(**kwargs):
        try:
            output = await pipe_method(f, **kwargs)
            await on_success()
            return output
        except HTTPResponse:
            await on_success()
            raise
        except Exception:
            await on_failure()
            raise

    return flow


def _wrap_flow_request_success(pipe_method, on_success, on_failure, f):
    @wraps(f)
    async def flow(**kwargs):
        try:
            output = await pipe_method(f, **kwargs)
            await on_success()
            return output
        except HTTPResponse:
            await on_success()
            raise

    return flow


def _wrap_flow_request_failure(pipe_method, on_success, on_failure, f):
    @wraps(f)
    async def flow(**kwargs):
        try:
            return await pipe_method(f, **kwargs)
        except HTTPResponse:
            raise
        except Exception:
            await on_failure()
            raise

    return flow


def _wrap_flow_request_basic(pipe_method, on_success, on_failure, f):
    @wraps(f)
    async def flow(**kwargs):
        return await pipe_method(f, **kwargs)

    return flow


def _wrap_flow_ws_complete(pipe_method, on_success, on_failure, f):
    @wraps(f)
    async def flow(**kwargs):
        try:
            await pipe_method(f, **kwargs)
            await on_success()
        except Exception:
            await on_failure()
            raise

    return flow


def _wrap_flow_ws_success(pipe_method, on_success, on_failure, f):
    @wraps(f)
    async def flow(**kwargs):
        await pipe_method(f, **kwargs)
        await on_success()

    return flow


def _wrap_flow_ws_failure(pipe_method, on_success, on_failure, f):
    @wraps(f)
    async def flow(**kwargs):
        try:
            await pipe_method(f, **kwargs)
        except Exception:
            await on_failure()
            raise

    return flow


def _wrap_flow_ws_basic(pipe_method, on_success, on_failure, f):
    @wraps(f)
    async def flow(**kwargs):
        return await pipe_method(f, **kwargs)

    return flow



import codecs
from enum import Enum

from .. import _emmett_core


class PBKDF2_HMAC(Enum):
    sha1 = _emmett_core.pbkdf2_sha1
    sha256 = _emmett_core.pbkdf2_sha256
    sha384 = _emmett_core.pbkdf2_sha384
    sha512 = _emmett_core.pbkdf2_sha512


def pbkdf2_bin(
    data: bytes,
    salt: bytes,
    iterations: int = 10000,
    keylen: int = 32,
    hash_algorithm: PBKDF2_HMAC = PBKDF2_HMAC.sha256,
) -> bytes:
    return hash_algorithm.value(data, salt, iterations, keylen)


def pbkdf2_hex(
    data: str, salt: str, iterations: int = 10000, keylen: int = 32, hash_algorithm: PBKDF2_HMAC = PBKDF2_HMAC.sha256
) -> str:
    return codecs.encode(
        pbkdf2_bin(
            data=data.encode("utf8"),
            salt=salt.encode("utf8"),
            iterations=iterations,
            keylen=keylen,
            hash_algorithm=hash_algorithm,
        ),
        "hex_codec",
    ).decode("utf8")



from emmett_core.cryptography.kdf import pbkdf2_bin


def test_pbkdf2():
    key = pbkdf2_bin(b"some password", b"a" * 16, 32)
    assert key == b"oQ\xbd-\xdb\x04\x85VT\x00\xc4\xcco\x8d\xc4~\xd3~(\x9e\xaa\xb8\x95\xacQ\xc4c\xa2\xbc1\x83*"



from __future__ import annotations

from typing import Any

from ..http.response import HTTPAsyncIterResponse, HTTPBytesResponse, HTTPIterResponse, HTTPResponse, HTTPStringResponse
from ..http.wrappers.response import Response
from .rules import HTTPRoutingRule


class MetaResponseBuilder:
    __slots__ = ["route"]

    def __init__(self, route: HTTPRoutingRule):
        self.route = route

    def __call__(self, output: Any, response: Response) -> HTTPResponse:
        return output


class EmptyResponseBuilder(MetaResponseBuilder):
    http_cls = HTTPResponse

    def __call__(self, output: Any, response: Response) -> HTTPResponse:
        return self.http_cls(response.status, headers=response.headers, cookies=response.cookies)


class BytesResponseBuilder(MetaResponseBuilder):
    http_cls = HTTPBytesResponse

    def __call__(self, output: Any, response: Response) -> HTTPBytesResponse:
        return self.http_cls(response.status, output, headers=response.headers, cookies=response.cookies)


class StringResponseBuilder(MetaResponseBuilder):
    http_cls = HTTPStringResponse

    def __call__(self, output: Any, response: Response) -> HTTPStringResponse:
        return self.http_cls(response.status, output, headers=response.headers, cookies=response.cookies)


class IterResponseBuilder(MetaResponseBuilder):
    http_cls = HTTPIterResponse

    def __call__(self, output: Any, response: Response) -> HTTPIterResponse:
        return self.http_cls(output, status_code=response.status, headers=response.headers, cookies=response.cookies)


class AsyncIterResponseBuilder(MetaResponseBuilder):
    http_cls = HTTPAsyncIterResponse

    def __call__(self, output: Any, response: Response) -> HTTPAsyncIterResponse:
        return self.http_cls(output, status_code=response.status, headers=response.headers, cookies=response.cookies)


class ResponseProcessor(StringResponseBuilder):
    def process(self, output: Any, response: Response):
        raise NotImplementedError

    def __call__(self, output: Any, response: Response) -> HTTPResponse:
        if isinstance(output, HTTPResponse):
            return output
        return self.http_cls(
            response.status, self.process(output, response), headers=response.headers, cookies=response.cookies
        )


class AutoResponseBuilder(ResponseProcessor):
    def process(self, output: Any, response: Response) -> str:
        if output is None:
            return ""
        if isinstance(output, str):
            return output
        return str(output)



from functools import partial
from typing import Any, Callable, Dict

from ._imports import orjson, rapidjson


if orjson is not None:
    _json_impl = orjson.loads
    _json_opts = {}
elif rapidjson is not None:
    _json_impl = rapidjson.loads
    _json_opts = {"datetime_mode": rapidjson.DM_ISO8601 | rapidjson.DM_NAIVE_IS_UTC, "number_mode": rapidjson.NM_NATIVE}
else:
    import json as _stdjson

    _json_impl = _stdjson.loads
    _json_opts = {}


class Parsers(object):
    _registry_: Dict[str, Callable[[str], Dict[str, Any]]] = {}

    @classmethod
    def register_for(cls, target):
        def wrap(f):
            cls._registry_[target] = f
            return f

        return wrap

    @classmethod
    def get_for(cls, target):
        return cls._registry_[target]


json = partial(_json_impl, **_json_opts)

Parsers.register_for("json")(json)



from __future__ import annotations

import os
import pkgutil
import sys
import traceback
import warnings
from types import ModuleType
from typing import Any, Generic, Optional

from .typing import T


class ProxyMixin:
    def _get_robj(self):
        raise NotImplementedError

    def __getitem__(self, key):
        return self._get_robj()[key]

    def __setitem__(self, key, value):
        self._get_robj()[key] = value

    def __delitem__(self, key):
        del self._get_robj()[key]

    def __getattr__(self, name: str) -> Any:
        return getattr(self._get_robj(), name)

    def __setattr__(self, name, value):
        setattr(self._get_robj(), name, value)

    def __delattr__(self, name):
        delattr(self._get_robj(), name)

    def __bool__(self):
        try:
            return bool(self._get_robj())
        except RuntimeError:
            return False

    def __eq__(self, obj) -> bool:
        return self._get_robj() == obj

    def __ne__(self, obj) -> bool:
        return self._get_robj() != obj

    def __call__(self, *args, **kwargs):
        return self._get_robj()(*args, **kwargs)

    def __iter__(self):
        return iter(self._get_robj())

    def __contains__(self, element):
        return element in self._get_robj()

    def __dir__(self):
        try:
            return dir(self._get_robj())
        except RuntimeError:
            return []

    @property
    def __dict__(self):
        try:
            return self._get_robj().__dict__
        except RuntimeError:
            raise AttributeError("__dict__")

    def __str__(self):
        return str(self._get_robj())

    def __repr__(self):
        try:
            obj = self._get_robj()
        except RuntimeError:
            return "<%s unbound>" % self.__class__.__name__
        return repr(obj)


class ObjectProxy(ProxyMixin, Generic[T]):
    __slots__ = ("__obj", "__name__")

    def __init__(self, obj: Any, name: str):
        object.__setattr__(self, "_ObjectProxy__obj", obj)
        object.__setattr__(self, "__name__", name)

    def _get_robj(self) -> T:
        return getattr(self.__obj, self.__name__)


class ContextVarProxy(ProxyMixin, Generic[T]):
    __slots__ = ("__obj", "__name__")

    def __init__(self, obj: Any, name: str):
        object.__setattr__(self, "_ContextVarProxy__obj", obj)
        object.__setattr__(self, "__name__", name)

    def _get_robj(self) -> T:
        return getattr(self.__obj.get(), self.__name__)


#: application loaders
def get_app_module(module_name: str, raise_on_failure: bool = True) -> Optional[ModuleType]:
    try:
        __import__(module_name)
    except ImportError:
        if sys.exc_info()[-1].tb_next:
            raise RuntimeError(
                f"While importing '{module_name}', an ImportError was raised:\n\n{traceback.format_exc()}"
            )
        elif raise_on_failure:
            raise RuntimeError(f"Could not import '{module_name}'.")
        else:
            return
    return sys.modules[module_name]


#: Given a module instance this tries to find the best possible application in the module
def find_best_app(module: ModuleType, app_cls) -> Any:
    # Search for the most common names first.
    for attr_name in ("app", "application"):
        app = getattr(module, attr_name, None)
        if isinstance(app, app_cls):
            return app

    # Otherwise find the only object that is an App instance.
    matches = [v for v in module.__dict__.values() if isinstance(v, app_cls)]

    if len(matches) == 1:
        return matches[0]
    raise RuntimeError(f"Failed to find Emmett application in module '{module.__name__}'.")


def locate_app(app_cls, module_name: str, app_name: str, raise_on_failure: bool = True) -> Any:
    module = get_app_module(module_name, raise_on_failure=raise_on_failure)
    if app_name:
        return getattr(module, app_name, None)
    return find_best_app(module, app_cls)


#: deprecation helpers
class RemovedInNextVersionWarning(DeprecationWarning):
    pass


class deprecated(object):
    def __init__(self, old_method_name, new_method_name, class_name=None, s=0):
        self.class_name = class_name
        self.old_method_name = old_method_name
        self.new_method_name = new_method_name
        self.additional_stack = s

    def __call__(self, f):
        def wrapped(*args, **kwargs):
            warn_of_deprecation(self.old_method_name, self.new_method_name, self.class_name, 3 + self.additional_stack)
            return f(*args, **kwargs)

        return wrapped


warnings.simplefilter("always", RemovedInNextVersionWarning)


def warn_of_deprecation(old_name, new_name, prefix=None, stack=2):
    msg = "%(old)s is deprecated, use %(new)s instead."
    if prefix:
        msg = "%(prefix)s." + msg
    warnings.warn(msg % {"old": old_name, "new": new_name, "prefix": prefix}, RemovedInNextVersionWarning, stack)


#: app init helpers
def get_root_path(import_name):
    """Returns the path of the package or cwd if that cannot be found."""
    # Module already imported and has a file attribute.  Use that first.
    mod = sys.modules.get(import_name)
    if mod is not None and hasattr(mod, "__file__"):
        return os.path.dirname(os.path.abspath(mod.__file__))

    # Next attempt: check the loader.
    loader = pkgutil.get_loader(import_name)

    # Loader does not exist or we're referring to an unloaded main module
    # or a main module without path (interactive sessions), go with the
    # current working directory.
    if loader is None or import_name == "__main__":
        return os.getcwd()

    # For .egg, zipimporter does not have get_filename until Python 2.7.
    # Some other loaders might exhibit the same behavior.
    if hasattr(loader, "get_filename"):
        filepath = loader.get_filename(import_name)
    else:
        # Fall back to imports.
        __import__(import_name)
        mod = sys.modules[import_name]
        filepath = getattr(mod, "__file__", None)

        # If we don't have a filepath it might be because we are a
        # namespace package.  In this case we pick the root path from the
        # first module that is contained in our package.
        if filepath is None:
            raise RuntimeError(
                "No root path can be found for the provided "
                'module "%s".  This can happen because the '
                "module came from an import hook that does "
                "not provide file name information or because "
                "it's a namespace package.  In this case "
                "the root path needs to be explicitly "
                "provided." % import_name
            )

    # filepath is import_name.py for a module, or __init__.py for a package.
    return os.path.dirname(os.path.abspath(filepath))


def create_missing_app_folders(app, folders=["static"]):
    try:
        for subfolder in folders:
            path = os.path.join(app.root_path, subfolder)
            if not os.path.exists(path):
                os.mkdir(path)
    except Exception:
        pass



import codecs
import mimetypes
import re
import sys
from http.cookiejar import CookieJar
from io import BytesIO
from urllib.request import Request as U2Request

from ....datastructures import sdict
from ._shortcuts import to_bytes, to_unicode
from .urls import get_host, uri_to_iri, url_quote


_quoted_string_re = r'"[^"\\]*(?:\\.[^"\\]*)*"'
_option_header_piece_re = re.compile(
    r";\s*(%s|[^\s;,=]+)\s*(?:=\s*(%s|[^;,]+)?)?\s*" % (_quoted_string_re, _quoted_string_re)
)
_option_header_start_mime_type = re.compile(r",\s*([^;,\s]+)([;,]\s*.+)?")


class _TestCookieHeaders:
    def __init__(self, headers):
        self.headers = headers

    def getheaders(self, name):
        headers = []
        name = name.lower()
        for k, v in self.headers:
            if k.lower() == name:
                headers.append(v)
        return headers

    def get_all(self, name, default=None):
        rv = []
        for k, v in self.headers:
            if k.lower() == name.lower():
                rv.append(v)
        return rv or default or []


class _TestCookieResponse:
    def __init__(self, headers):
        self.headers = _TestCookieHeaders(headers)

    def info(self):
        return self.headers


class TestCookieJar(CookieJar):
    def inject_rsgi(self, scope):
        cvals = []
        for cookie in self:
            cvals.append("%s=%s" % (cookie.name, cookie.value))
        if cvals:
            scope.headers._inject([("cookie", "; ".join(cvals))])

    def extract_rsgi(self, scope, headers):
        self.extract_cookies(
            _TestCookieResponse(headers),
            U2Request(get_current_url(scope)),  # noqa: S310
        )


class Headers(dict):
    def __init__(self, headers=[]):
        super(Headers, self).__init__()
        self._list = []
        for header in headers:
            key, value = header[0].lower(), header[1]
            self._list.append((key, value))
            self[key] = value

    def __getitem__(self, name):
        return super(Headers, self).__getitem__(name.lower())

    def __setitem__(self, name, value):
        return super(Headers, self).__setitem__(name.lower(), value)

    def __iter__(self):
        return iter(self._list)

    def get(self, name, d=None, type=None):
        rv = super(Headers, self).get(name.lower)
        if rv is None:
            return d
        if type is None:
            return rv
        try:
            rv = type(rv)
        except ValueError:
            pass
        return rv


class _FileHandler(object):
    def __init__(self, stream=None, filename=None, name=None, content_type=None, content_length=None, headers=None):
        self.name = name
        self.stream = stream or BytesIO()

        if filename is None:
            filename = getattr(stream, "name", None)
            if filename and filename[0] == "<" and filename[-1] == ">":
                filename = None

            if isinstance(filename, bytes):
                filename = filename.decode(get_filesystem_encoding(), "replace")

        self.filename = filename
        if headers is None:
            headers = Headers()
        self.headers = headers
        if content_type is not None:
            headers["Content-Type"] = content_type
        if content_length is not None:
            headers["Content-Length"] = str(content_length)

    def _parse_content_type(self):
        if not hasattr(self, "_parsed_content_type"):
            self._parsed_content_type = parse_options_header(self.content_type)

    @property
    def content_type(self):
        return self.headers.get("content-type")

    @property
    def content_length(self):
        return int(self.headers.get("content-length") or 0)

    @property
    def mimetype(self):
        self._parse_content_type()
        return self._parsed_content_type[0].lower()

    @property
    def mimetype_params(self):
        self._parse_content_type()
        return self._parsed_content_type[1]

    def save(self, dst, buffer_size=16384):
        from shutil import copyfileobj

        close_dst = False
        if isinstance(dst, str):
            dst = open(dst, "wb")
            close_dst = True
        try:
            copyfileobj(self.stream, dst, buffer_size)
        finally:
            if close_dst:
                dst.close()

    def close(self):
        try:
            self.stream.close()
        except Exception:
            pass

    def __nonzero__(self):
        return bool(self.filename)

    __bool__ = __nonzero__

    def __getattr__(self, name):
        return getattr(self.stream, name)

    def __iter__(self):
        return iter(self.readline, "")

    def __repr__(self):
        return "<%s: %r (%r)>" % (self.__class__.__name__, self.filename, self.content_type)


class filesdict(sdict):
    def add_file(self, name, file, filename=None, content_type=None):
        if isinstance(file, _FileHandler):
            value = file
        else:
            if isinstance(file, str):
                if filename is None:
                    filename = file
                file = open(file, "rb")
            if filename and content_type is None:
                content_type = mimetypes.guess_type(filename)[0] or "application/octet-stream"
            value = _FileHandler(file, filename, name, content_type)
        if name not in self:
            self[name] = []
        self[name].append(value)


def get_current_url(scope, root_only=False, strip_querystring=False, host_only=False):
    headers = Headers(list(scope.headers._denorm()))
    components = [scope.scheme, "://", get_host(scope, headers)]
    if host_only:
        return uri_to_iri("".join(components) + "/")
    components.extend([url_quote(to_bytes("/")).rstrip("/"), "/"])
    if not root_only:
        components.append(url_quote(to_bytes(scope.path).lstrip(b"/")))
        if not strip_querystring:
            if scope.query_string:
                components.extend(["?", to_unicode(scope.query_string)])
    return uri_to_iri("".join(components))


def _is_ascii_encoding(encoding):
    if encoding is None:
        return False
    try:
        return codecs.lookup(encoding).name == "ascii"
    except LookupError:
        return False


def get_filesystem_encoding():
    rv = sys.getfilesystemencoding()
    if (sys.platform.startswith("linux") or "bsd" in sys.platform) and not rv or _is_ascii_encoding(rv):
        return "utf-8"
    return rv


def unquote_header_value(value, is_filename=False):
    if value and value[0] == value[-1] == '"':
        value = value[1:-1]
        if not is_filename or value[:2] != "\\\\":
            return value.replace("\\\\", "\\").replace('\\"', '"')
    return value


def parse_options_header(value, multiple=False):
    if not value:
        return "", {}
    result = []
    value = "," + value.replace("\n", ",")
    while value:
        match = _option_header_start_mime_type.match(value)
        if not match:
            break
        result.append(match.group(1))  # mimetype
        options = {}
        # Parse options
        rest = match.group(2)
        while rest:
            optmatch = _option_header_piece_re.match(rest)
            if not optmatch:
                break
            option, option_value = optmatch.groups()
            option = unquote_header_value(option)
            if option_value is not None:
                option_value = unquote_header_value(option_value, option == "filename")
            options[option] = option_value
            rest = rest[optmatch.end() :]
        result.append(options)
        if multiple is False:
            return tuple(result)
        value = rest

    return tuple(result)


def stream_encode_multipart(values, threshold=1024 * 500, boundary=None, charset="utf-8"):
    if boundary is None:
        from random import random
        from time import time

        boundary = "---------------EmmettFormPart_%s%s" % (time(), random())  # noqa: S311
    _closure = [BytesIO(), 0, False]

    write_binary = _closure[0].write

    def write(string):
        write_binary(string.encode(charset))

    for key, values in values.items():  # noqa: B020
        if not isinstance(values, list):
            values = [values]
        for value in values:
            write('--%s\r\nContent-Disposition: form-data; name="%s"' % (boundary, key))
            reader = getattr(value, "read", None)
            if reader is not None:
                filename = getattr(value, "filename", getattr(value, "name", None))
                content_type = getattr(value, "content_type", None)
                if content_type is None:
                    content_type = filename and mimetypes.guess_type(filename)[0] or "application/octet-stream"
                if filename is not None:
                    write('; filename="%s"\r\n' % filename)
                else:
                    write("\r\n")
                write("Content-Type: %s\r\n\r\n" % content_type)
                while 1:
                    chunk = reader(16384)
                    if not chunk:
                        break
                    write_binary(chunk)
            else:
                if not isinstance(value, str):
                    value = str(value)
                value = to_bytes(value, charset)
                write("\r\n\r\n")
                write_binary(value)
            write("\r\n")
    write("--%s--\r\n" % boundary)

    length = int(_closure[0].tell())
    _closure[0].seek(0)
    return _closure[0], length, boundary



from emmett_core.cryptography.ciphers import AESModes, aes_decrypt, aes_encrypt


text = b"plain text"


def test_aes128_ctr():
    key = b"a" * 16
    iv = b"b" * 16

    ct = aes_encrypt(text, key, iv, AESModes.CTR128)
    assert ct == b"i%\xd4\xcd\x1ew6R\xa9\xf3"
    assert aes_decrypt(ct, key, iv, AESModes.CTR128) == text


def test_aes256_ctr():
    key = b"a" * 32
    iv = b"b" * 16

    ct = aes_encrypt(text, key, iv, AESModes.CTR128)
    assert ct == b"\xba\x18&\xe2\x81\x1b\x85Xr\xbb"
    assert aes_decrypt(ct, key, iv, AESModes.CTR128) == text


def test_aes128_cfb8():
    key = b"a" * 16
    iv = b"b" * 16

    ct = aes_encrypt(text, key, iv, AESModes.CFB8)
    assert ct == b"i\x9d\xe6\xc1\xefTC\xd0\xb1\x8e"
    assert aes_decrypt(ct, key, iv, AESModes.CFB8) == text


def test_aes256_cfb8():
    key = b"a" * 32
    iv = b"b" * 16

    ct = aes_encrypt(text, key, iv, AESModes.CFB8)
    assert ct == b'\xba\xe1"\x85\x0f\t\x84\x91\xb2E'
    assert aes_decrypt(ct, key, iv, AESModes.CFB8) == text


def test_aes128_cfb128():
    key = b"a" * 16
    iv = b"b" * 16

    ct = aes_encrypt(text, key, iv, AESModes.CFB128)
    assert ct == b"i%\xd4\xcd\x1ew6R\xa9\xf3"
    assert aes_decrypt(ct, key, iv, AESModes.CFB128) == text


def test_aes256_cfb128():
    key = b"a" * 32
    iv = b"b" * 16

    ct = aes_encrypt(text, key, iv, AESModes.CFB128)
    assert ct == b"\xba\x18&\xe2\x81\x1b\x85Xr\xbb"
    assert aes_decrypt(ct, key, iv, AESModes.CFB128) == text



import copy
from typing import Dict, Optional

from .typing import KT, VT


class sdict(Dict[KT, VT]):
    #: like a dictionary except `obj.foo` can be used in addition to
    #  `obj['foo']`, and setting obj.foo = None deletes item foo.
    __slots__ = ()

    __setattr__ = dict.__setitem__  # type: ignore
    __delattr__ = dict.__delitem__  # type: ignore
    __getitem__ = dict.get  # type: ignore

    # see http://stackoverflow.com/questions/10364332/how-to-pickle-python-object-derived-from-dict
    def __getattr__(self, key: str) -> Optional[VT]:
        if key.startswith("__"):
            raise AttributeError
        return self.get(key, None)  # type: ignore

    __repr__ = lambda self: "<sdict %s>" % dict.__repr__(self)
    __getstate__ = lambda self: None
    __copy__ = lambda self: sdict(self)
    __deepcopy__ = lambda self, memo: sdict(copy.deepcopy(dict(self)))


class gsdict(sdict[KT, VT]):
    #: like sdict, except it autogrows creating sub-sdict attributes
    __slots__ = ()

    def __getitem__(self, key):
        if key not in self.keys():
            self[key] = sdict()
        return super().__getitem__(key)

    __getattr__ = __getitem__


class ImmutableListMixin:
    _hash_cache = None

    def __hash__(self) -> Optional[int]:  # type: ignore
        if self._hash_cache is not None:
            return self._hash_cache
        rv = self._hash_cache = hash(tuple(self))  # type: ignore
        return rv

    def __reduce_ex__(self, protocol):
        return type(self), (list(self),)

    def __delitem__(self, key):
        _is_immutable(self)

    def __iadd__(self, other):
        _is_immutable(self)

    def __imul__(self, other):
        _is_immutable(self)

    def __setitem__(self, key, value):
        _is_immutable(self)

    def append(self, item):
        _is_immutable(self)

    def remove(self, itme):
        _is_immutable(self)

    def extend(self, iterable):
        _is_immutable(self)

    def insert(self, pos, value):
        _is_immutable(self)

    def pop(self, index=-1):
        _is_immutable(self)

    def reverse(self):
        _is_immutable(self)

    def sort(self, cmp=None, key=None, reverse=None):
        _is_immutable(self)


class ImmutableList(ImmutableListMixin, list):  # type: ignore
    def __repr__(self):
        return "%s(%s)" % (self.__class__.__name__, list.__repr__(self))


def _is_immutable(self):
    raise TypeError("%r objects are immutable" % self.__class__.__name__)



import sys
from io import BytesIO
from urllib.parse import parse_qs

from ....datastructures import sdict
from .helpers import Headers, filesdict, stream_encode_multipart
from .urls import iri_to_uri, url_encode, url_fix, url_parse, url_unparse


class ScopeHeaders:
    def __init__(self, data):
        self.inner = {}
        self._inject(data)

    def _inject(self, data):
        for key, val in data:
            if key in self.inner:
                self.inner[key].append(val)
            else:
                self.inner[key] = [val]

    def __contains__(self, key: str):
        return key in self.inner

    def keys(self):
        return self.inner.keys()

    def values(self):
        for key in self.inner.keys():
            yield self.inner[key][0]

    def items(self):
        for key in self.inner.keys():
            yield key, self.inner[key][0]

    def get(self, key, default=None):
        if rv := self.inner.get(key):
            return rv[0]
        return default

    def get_all(self, key, default=None):
        return self.inner.get(key, default)

    def _denorm(self):
        for key in self.inner.keys():
            for val in self.inner[key]:
                yield key, val


class Scope:
    def __init__(
        self,
        headers,
        path,
        method,
        query_string,
        scheme="http",
        proto="http",
        http_version="1.1",
        rsgi_version="2.4",
        server="127.0.0.1",
        client="127.0.0.1",
        authority=None,
    ):
        self.headers = ScopeHeaders(headers)
        self.path = path
        self.method = method
        self.query_string = query_string
        # self.body = body
        self.scheme = scheme
        self.proto = proto
        self.http_version = http_version
        self.rsgi_version = rsgi_version
        self.server = server
        self.client = client
        self.authority = authority


class ScopeBuilder(object):
    server_protocol = "1.1"

    def __init__(
        self,
        path="/",
        base_url=None,
        query_string=None,
        method="GET",
        input_stream=None,
        content_type=None,
        content_length=None,
        errors_stream=None,
        headers=None,
        data=None,
        charset="utf-8",
        boundary=None,
    ):
        if query_string is None and "?" in path:
            path, query_string = path.split("?", 1)
        self.charset = charset
        self.path = iri_to_uri(path)
        if base_url is not None:
            base_url = url_fix(iri_to_uri(base_url, charset), charset)
        self.base_url = base_url
        if isinstance(query_string, (bytes, str)):
            self.query_string = query_string
        else:
            if query_string is None:
                query_string = sdict()
            elif not isinstance(query_string, dict):
                query_string = self._parse_querystring(query_string)
            self.args = query_string
        self.method = method
        if headers is None:
            headers = Headers()
        elif not isinstance(headers, Headers):
            headers = Headers(headers)
        self.headers = headers
        if content_type is not None:
            self.content_type = content_type
        if errors_stream is None:
            errors_stream = sys.stderr
        self.errors_stream = errors_stream
        self.input_stream = input_stream
        self.content_length = content_length
        self.boundary = boundary
        self.closed = False

        if data:
            if input_stream is not None:
                raise TypeError("can't provide input stream and data")
            if isinstance(data, str):
                data = data.encode(self.charset)
            if isinstance(data, bytes):
                self.input_stream = BytesIO(data)
                if self.content_length is None:
                    self.content_length = len(data)
            else:
                for key, values in data.items():
                    if not isinstance(values, list):
                        values = [values]
                    for v in values:
                        if isinstance(v, (tuple)) or hasattr(v, "read"):
                            self._add_file_from_data(key, v)
                        else:
                            if self.form[key] is None:
                                self.form[key] = []
                            self.form[key].append(v)

    @staticmethod
    def _parse_querystring(query_string):
        dget = parse_qs(query_string, keep_blank_values=1)
        params = sdict(dget)
        for key, value in params.items():
            if isinstance(value, list) and len(value) == 1:
                params[key] = value[0]
        return params

    def _add_file_from_data(self, key, value):
        if isinstance(value, tuple):
            self.files.add_file(key, *value)
        else:
            self.files.add_file(key, value)

    def _get_base_url(self):
        return url_unparse((self.url_scheme, self.host, self.script_root, "", "")).rstrip("/") + "/"

    def _set_base_url(self, value):
        if value is None:
            scheme = "http"
            netloc = "localhost"
            script_root = ""
        else:
            scheme, netloc, script_root, qs, anchor = url_parse(value)
            if qs or anchor:
                raise ValueError("base url must not contain a query string or fragment")
        self.script_root = script_root.rstrip("/")
        self.host = netloc
        self.url_scheme = scheme

    base_url = property(_get_base_url, _set_base_url)
    del _get_base_url, _set_base_url

    def _get_content_type(self):
        ct = self.headers.get("Content-Type")
        if ct is None and not self._input_stream:
            if self._files:
                return "multipart/form-data"
            elif self._form:
                return "application/x-www-form-urlencoded"
            return None
        return ct

    def _set_content_type(self, value):
        if value is None:
            self.headers.pop("Content-Type", None)
        else:
            self.headers["Content-Type"] = value

    content_type = property(_get_content_type, _set_content_type)
    del _get_content_type, _set_content_type

    def _get_content_length(self):
        return self.headers.get("Content-Length", type=int)

    def _set_content_length(self, value):
        if value is None:
            self.headers.pop("Content-Length", None)
        else:
            self.headers["Content-Length"] = str(value)

    content_length = property(_get_content_length, _set_content_length)
    del _get_content_length, _set_content_length

    def form_property(name, storage):  # noqa: N805
        key = "_" + name

        def getter(self):
            if self._input_stream is not None:
                raise AttributeError("an input stream is defined")
            rv = getattr(self, key)
            if rv is None:
                rv = storage()
                setattr(self, key, rv)

            return rv

        def setter(self, value):
            self._input_stream = None
            setattr(self, key, value)

        return property(getter, setter)

    form = form_property("form", sdict)
    files = form_property("files", filesdict)
    del form_property

    def _get_input_stream(self):
        return self._input_stream

    def _set_input_stream(self, value):
        self._input_stream = value
        self._form = self._files = None

    input_stream = property(
        _get_input_stream,
        _set_input_stream,
    )
    del _get_input_stream, _set_input_stream

    def _get_query_string(self):
        if self._query_string is None:
            if self._args is not None:
                return url_encode(self._args, charset=self.charset)
            return ""
        return self._query_string

    def _set_query_string(self, value):
        self._query_string = value
        self._args = None

    query_string = property(_get_query_string, _set_query_string)
    del _get_query_string, _set_query_string

    def _get_args(self):
        if self._query_string is not None:
            raise AttributeError("a query string is defined")
        if self._args is None:
            self._args = sdict()
        return self._args

    def _set_args(self, value):
        self._query_string = None
        self._args = value

    args = property(_get_args, _set_args)
    del _get_args, _set_args

    @property
    def server_name(self):
        return self.host.split(":", 1)[0]

    @property
    def server_port(self):
        pieces = self.host.split(":", 1)
        if len(pieces) == 2 and pieces[1].isdigit():
            return int(pieces[1])
        elif self.url_scheme == "https":
            return 443
        return 80

    def __del__(self):
        try:
            self.close()
        except Exception:
            pass

    def close(self):
        if self.closed:
            return
        try:
            files = self.files.values()
        except AttributeError:
            files = ()
        for f in files:
            try:
                f.close()
            except Exception:
                pass
        self.closed = True

    def get_data(self):
        input_stream = self.input_stream
        content_length = self.content_length
        content_type = self.content_type

        if input_stream is not None:
            start_pos = input_stream.tell()
            input_stream.seek(0, 2)
            end_pos = input_stream.tell()
            input_stream.seek(start_pos)
            content_length = end_pos - start_pos
        elif content_type == "multipart/form-data":
            values = sdict()
            for d in [self.files, self.form]:
                for key, val in d.items():
                    if key not in values:
                        values[key] = []
                    values[key].extend(val)
            input_stream, content_length, boundary = stream_encode_multipart(
                values, charset=self.charset, boundary=self.boundary
            )
            content_type += '; boundary="%s"' % boundary
        elif content_type == "application/x-www-form-urlencoded":
            values = url_encode(self.form, charset=self.charset)
            values = values.encode("ascii")
            content_length = len(values)
            input_stream = BytesIO(values)
        else:
            input_stream = BytesIO()

        headers = [("host", self.host)]
        if self.content_type:
            headers.extend([("content-type", content_type), ("content-length", str(content_length or 0))])
        for key, value in self.headers:
            headers.append((key, value))

        scope = Scope(
            headers=headers,
            path=self.path,
            method=self.method,
            query_string=self.query_string,
            scheme=self.url_scheme,
            http_version=self.server_protocol,
            server=f"{self.server_name}:{self.server_port}",
            client=f"{self.server_name}:{self.server_port}",
        )

        return scope, input_stream.getvalue()



from __future__ import annotations

import errno
import mimetypes
import os
import stat
from email.utils import formatdate
from hashlib import md5
from typing import Any, AsyncIterable, BinaryIO, Dict, Generator, Iterable, Tuple

from .._io import loop_open_file


class HTTPResponse(Exception):
    def __init__(
        self,
        status_code: int,
        *,
        headers: Dict[str, str] = {"content-type": "text/plain"},
        cookies: Dict[str, Any] = {},
    ):
        self.status_code: int = status_code
        self._headers: Dict[str, str] = headers
        self._cookies: Dict[str, Any] = cookies

    def asgi_headers(self) -> Generator[Tuple[bytes, bytes], None, None]:
        for key, val in self._headers.items():
            yield key.encode("latin-1"), val.encode("latin-1")
        for cookie in self._cookies.values():
            yield b"set-cookie", str(cookie)[12:].encode("latin-1")

    def rsgi_headers(self) -> Generator[Tuple[str, str], None, None]:
        for key, val in self._headers.items():
            yield key, val
        for cookie in self._cookies.values():
            yield "set-cookie", str(cookie)[12:]

    async def _send_headers(self, send):
        await send({"type": "http.response.start", "status": self.status_code, "headers": list(self.asgi_headers())})

    async def _send_body(self, send):
        await send({"type": "http.response.body"})

    async def asgi(self, scope, send):
        await self._send_headers(send)
        await self._send_body(send)

    def rsgi(self, protocol):
        protocol.response_empty(self.status_code, list(self.rsgi_headers()))


class HTTPBytesResponse(HTTPResponse):
    def __init__(
        self,
        status_code: int,
        body: bytes = b"",
        headers: Dict[str, str] = {"content-type": "text/plain"},
        cookies: Dict[str, Any] = {},
    ):
        super().__init__(status_code, headers=headers, cookies=cookies)
        self.body = body

    async def _send_body(self, send):
        await send({"type": "http.response.body", "body": self.body, "more_body": False})

    def rsgi(self, protocol):
        protocol.response_bytes(self.status_code, list(self.rsgi_headers()), self.body)


class HTTPStringResponse(HTTPResponse):
    def __init__(
        self,
        status_code: int,
        body: str = "",
        headers: Dict[str, str] = {"content-type": "text/plain"},
        cookies: Dict[str, Any] = {},
    ):
        super().__init__(status_code, headers=headers, cookies=cookies)
        self.body = body

    @property
    def encoded_body(self):
        return self.body.encode("utf-8")

    async def _send_body(self, send):
        await send({"type": "http.response.body", "body": self.encoded_body, "more_body": False})

    def rsgi(self, protocol):
        protocol.response_str(self.status_code, list(self.rsgi_headers()), self.body)


class HTTPRedirectResponse(HTTPResponse):
    def __init__(self, status_code: int, location: str, cookies: Dict[str, Any] = {}):
        location = location.replace("\r", "%0D").replace("\n", "%0A")
        super().__init__(status_code, headers={"location": location}, cookies=cookies)


class HTTPFileResponse(HTTPResponse):
    def __init__(
        self,
        file_path: str,
        status_code: int = 200,
        headers: Dict[str, str] = {},
        cookies: Dict[str, Any] = {},
        chunk_size: int = 4096,
    ):
        super().__init__(status_code, headers=headers, cookies=cookies)
        self.file_path = file_path
        self.chunk_size = chunk_size

    def _get_stat_headers(self, stat_data):
        content_type = mimetypes.guess_type(self.file_path)[0] or "text/plain"
        content_length = str(stat_data.st_size)
        last_modified = formatdate(stat_data.st_mtime, usegmt=True)
        etag_base = str(stat_data.st_mtime) + "_" + str(stat_data.st_size)
        etag = md5(etag_base.encode("utf-8")).hexdigest()  # noqa: S324
        return {
            "content-type": content_type,
            "content-length": content_length,
            "last-modified": last_modified,
            "etag": etag,
        }

    async def asgi(self, scope, send):
        try:
            stat_data = os.stat(self.file_path)
            if not stat.S_ISREG(stat_data.st_mode):
                await HTTPResponse(403).send(scope, send)
                return
            self._headers.update(self._get_stat_headers(stat_data))
            await self._send_headers(send)
            if "http.response.pathsend" in scope.get("extensions", {}):
                await send({"type": "http.response.pathsend", "path": str(self.file_path)})
            else:
                await self._send_body(send)
        except IOError as e:
            if e.errno == errno.EACCES:
                await HTTPResponse(403).send(scope, send)
            else:
                await HTTPResponse(404).send(scope, send)

    async def _send_body(self, send):
        async with loop_open_file(self.file_path, mode="rb") as f:
            more_body = True
            while more_body:
                chunk = await f.read(self.chunk_size)
                more_body = len(chunk) == self.chunk_size
                await send(
                    {
                        "type": "http.response.body",
                        "body": chunk,
                        "more_body": more_body,
                    }
                )

    def rsgi(self, protocol):
        try:
            stat_data = os.stat(self.file_path)
            if not stat.S_ISREG(stat_data.st_mode):
                return HTTPResponse(403).rsgi(protocol)
            self._headers.update(self._get_stat_headers(stat_data))
        except IOError as e:
            if e.errno == errno.EACCES:
                return HTTPResponse(403).rsgi(protocol)
            return HTTPResponse(404).rsgi(protocol)

        protocol.response_file(self.status_code, list(self.rsgi_headers()), self.file_path)


class HTTPIOResponse(HTTPResponse):
    def __init__(
        self,
        io_stream: BinaryIO,
        status_code: int = 200,
        headers: Dict[str, str] = {},
        cookies: Dict[str, Any] = {},
        chunk_size: int = 4096,
    ):
        super().__init__(status_code, headers=headers, cookies=cookies)
        self.io_stream = io_stream
        self.chunk_size = chunk_size

    def _get_io_headers(self):
        content_length = str(self.io_stream.getbuffer().nbytes)
        return {"content-length": content_length}

    async def asgi(self, scope, send):
        self._headers.update(self._get_io_headers())
        await self._send_headers(send)
        await self._send_body(send)

    async def _send_body(self, send):
        more_body = True
        while more_body:
            chunk = self.io_stream.read(self.chunk_size)
            more_body = len(chunk) == self.chunk_size
            await send(
                {
                    "type": "http.response.body",
                    "body": chunk,
                    "more_body": more_body,
                }
            )

    def rsgi(self, protocol):
        protocol.response_bytes(self.status_code, list(self.rsgi_headers()), self.io_stream.read())


class HTTPIterResponse(HTTPResponse):
    def __init__(
        self, iter: Iterable[bytes], status_code: int = 200, headers: Dict[str, str] = {}, cookies: Dict[str, Any] = {}
    ):
        super().__init__(status_code, headers=headers, cookies=cookies)
        self.iter = iter

    async def _send_body(self, send):
        for chunk in self.iter:
            await send({"type": "http.response.body", "body": chunk, "more_body": True})
        await send({"type": "http.response.body", "body": b"", "more_body": False})

    async def rsgi(self, protocol):
        trx = protocol.response_stream(self.status_code, list(self.rsgi_headers()))
        for chunk in self.iter:
            await trx.send_bytes(chunk)


class HTTPAsyncIterResponse(HTTPResponse):
    def __init__(
        self,
        iter: AsyncIterable[bytes],
        status_code: int = 200,
        headers: Dict[str, str] = {},
        cookies: Dict[str, Any] = {},
    ):
        super().__init__(status_code, headers=headers, cookies=cookies)
        self.iter = iter

    async def _send_body(self, send):
        async for chunk in self.iter:
            await send({"type": "http.response.body", "body": chunk, "more_body": True})
        await send({"type": "http.response.body", "body": b"", "more_body": False})

    async def rsgi(self, protocol):
        trx = protocol.response_stream(self.status_code, list(self.rsgi_headers()))
        async for chunk in self.iter:
            await trx.send_bytes(chunk)



import pytest

from emmett_core.sessions import SessionManager as _SessionManager


@pytest.fixture(scope="function")
def session_manager(current):
    class SessionManager(_SessionManager):
        @classmethod
        def _build_pipe(cls, handler_cls, *args, **kwargs):
            cls._pipe = handler_cls(current, *args, **kwargs)
            return cls._pipe

    return SessionManager.cookies(key="sid", secure=True, domain="localhost", cookie_name="foo_session")


@pytest.mark.asyncio
async def test_session_cookie(http_ctx, session_manager):
    assert session_manager.key == "sid"
    assert session_manager.secure is True
    assert session_manager.domain == "localhost"

    await session_manager.open_request()
    assert http_ctx.session._expiration == 3600

    await session_manager.close_request()
    cookie = str(http_ctx.response.cookies)
    assert "foo_session" in cookie
    assert "Domain=localhost;" in cookie
    assert "secure" in cookie.lower()

    http_ctx.request.cookies = http_ctx.response.cookies
    await session_manager.open_request()
    assert http_ctx.session._expiration == 3600



from contextlib import contextmanager

import pytest

from emmett_core.ctx import Current, RequestContext, WSContext
from emmett_core.datastructures import sdict
from emmett_core.routing.router import HTTPRouter, WebsocketRouter


@pytest.fixture(scope="function")
def app():
    return sdict(
        root_path="",
        languages=["en", "it"],
        language_default="en",
        language_force_on_url=True,
        send_signal=lambda *a, **k: None,
        config=sdict(
            hostname_default=None,
            static_version=None,
            static_version_urls=False,
        ),
    )


@pytest.fixture(scope="function")
def current(app):
    rv = Current()
    rv.app = app
    return rv


@pytest.fixture(scope="function")
def http_router(app, current):
    rv = HTTPRouter(app, current)
    app._router_http = rv
    return rv


@pytest.fixture(scope="function")
def ws_router(app, current):
    rv = WebsocketRouter(app, current)
    app._router_ws = rv
    return rv


@pytest.fixture(scope="function")
def http_ctx_builder(current):
    @contextmanager
    def ctx_builder(path, method="GET", scheme="http", host="localhost"):
        req = sdict(path=path, method=method, scheme=scheme, host=host)
        token = current._init_(RequestContext(current.app, req, sdict()))
        yield sdict(wrapper=current.request)
        current._close_(token)

    return ctx_builder


@pytest.fixture(scope="function")
def ws_ctx_builder(current):
    @contextmanager
    def ctx_builder(path, scheme="http", host="localhost"):
        ws = sdict(path=path, scheme=scheme, host=host)
        token = current._init_(WSContext(current.app, ws))
        yield sdict(wrapper=current.websocket)
        current._close_(token)

    return ctx_builder



from collections import defaultdict

import pytest

from emmett_core.cache import Cache
from emmett_core.cache.handlers import CacheHandler, RamCache


class RamCacheCustom(RamCache): ...


def test_basecache():
    base_cache = CacheHandler()
    assert base_cache._default_expire == 300

    assert base_cache("key", lambda: "value") == "value"
    assert base_cache.get("key") is None
    assert base_cache.set("key", "value", 300) is None
    assert base_cache.clear() is None


async def _await_2():
    return 2


async def _await_3():
    return 3


@pytest.mark.asyncio
async def test_ramcache():
    ram_cache = RamCache()
    assert ram_cache._prefix == ""
    assert ram_cache._threshold == 500

    assert ram_cache("test", lambda: 2) == 2
    assert ram_cache("test", lambda: 3, 300) == 2

    assert await ram_cache("test_loop", _await_2) == 2
    assert await ram_cache("test_loop", _await_3, 300) == 2

    ram_cache.set("test", 3)
    assert ram_cache.get("test") == 3

    ram_cache.set("test", 4, 300)
    assert ram_cache.get("test") == 4

    ram_cache.clear()
    assert ram_cache.get("test") is None


@pytest.mark.asyncio
async def test_cache():
    default_cache = Cache()
    assert isinstance(default_cache._default_handler, RamCache)

    assert default_cache("test", lambda: 2) == 2
    assert default_cache("test", lambda: 3) == 2

    assert await default_cache("test_loop", _await_2) == 2
    assert await default_cache("test_loop", _await_3, 300) == 2

    default_cache.set("test", 3)
    assert default_cache("test", lambda: 2) == 3

    default_cache.set("test", 4, 300)
    assert default_cache("test", lambda: 2, 300) == 4

    default_cache.clear()

    custom_cache = RamCacheCustom()
    ram_cache = RamCache()
    cache = Cache(default="custom", ram=ram_cache, custom=custom_cache)
    assert isinstance(cache._default_handler, RamCacheCustom)
    assert cache.custom == custom_cache
    assert cache.ram == ram_cache


def test_cache_decorator_sync():
    cache = Cache(ram=RamCache(prefix="test:"))
    calls = defaultdict(lambda: 0)

    @cache("foo")
    def foo(*args, **kwargs):
        calls["foo"] += 1
        return "foo"

    #: no arguments
    for _ in range(0, 2):
        foo()
    assert len(cache._default_handler.data.keys()) == 1
    assert calls["foo"] == 1

    #: args change the cache key
    for _ in range(0, 2):
        foo(1)
    assert len(cache._default_handler.data.keys()) == 2
    assert calls["foo"] == 2

    for _ in range(0, 2):
        foo(1, 2)
    assert len(cache._default_handler.data.keys()) == 3
    assert calls["foo"] == 3

    #: kwargs change the cache key
    for _ in range(0, 2):
        foo(1, a="foo", b="bar")
    assert len(cache._default_handler.data.keys()) == 4
    assert calls["foo"] == 4

    #: kwargs order won't change the cache key
    for _ in range(0, 2):
        foo(1, b="bar", a="foo")
    assert len(cache._default_handler.data.keys()) == 4
    assert calls["foo"] == 4


@pytest.mark.asyncio
async def test_cache_decorator_loop():
    cache = Cache(ram=RamCache(prefix="bar:"))
    calls = defaultdict(lambda: 0)

    @cache("bar")
    async def bar(*args, **kwargs):
        calls["bar"] += 1
        return "bar"

    #: no arguments
    for _ in range(0, 2):
        await bar()
    assert len(cache._default_handler.data.keys()) == 1
    assert calls["bar"] == 1

    #: args change the cache key
    for _ in range(0, 2):
        await bar(1)
    assert len(cache._default_handler.data.keys()) == 2
    assert calls["bar"] == 2

    for _ in range(0, 2):
        await bar(1, 2)
    assert len(cache._default_handler.data.keys()) == 3
    assert calls["bar"] == 3

    #: kwargs change the cache key
    for _ in range(0, 2):
        await bar(1, a="foo", b="bar")
    assert len(cache._default_handler.data.keys()) == 4
    assert calls["bar"] == 4

    #: kwargs order won't change the cache key
    for _ in range(0, 2):
        await bar(1, b="bar", a="foo")
    assert len(cache._default_handler.data.keys()) == 4
    assert calls["bar"] == 4



from __future__ import annotations

import re
from collections import namedtuple
from functools import partial
from typing import Any, Callable, List, Type

from .._emmett_core import HTTPRouter as _HTTPRouter, WSRouter as _WSRouter
from ..extensions import Signals
from ..http.response import HTTPBytesResponse
from .response import (
    AsyncIterResponseBuilder,
    AutoResponseBuilder,
    BytesResponseBuilder,
    EmptyResponseBuilder,
    IterResponseBuilder,
    MetaResponseBuilder,
    StringResponseBuilder,
)
from .rules import HTTPRoutingRule, RoutingRule, WebsocketRoutingRule


RouteRecReq = namedtuple("RouteRecReq", ["name", "dispatch", "flow_stream"])
RouteRecWS = namedtuple("RouteRecWS", ["name", "dispatch", "flow_recv", "flow_send"])


class RouterMixin:
    _routing_signal = Signals.before_routes
    _routing_started = False
    _routing_stack: List[RoutingRule] = []
    _re_components = re.compile(r"(\()?([^<\w]+)?<(\w+)\:(\w+)>(\)\?)?")

    @classmethod
    def _init_router_(cls, router, app, current, url_prefix=None):
        router.app = app
        router.current = current
        router.static_versioning = partial(cls.static_versioning, router)
        router.routes = []
        router.routes_out = {}
        router._routes_str = {}
        router._impl_host_match = False
        router._impl_scheme_match = False
        main_prefix = url_prefix or ""
        if main_prefix:
            main_prefix = main_prefix.rstrip("/")
            if not main_prefix.startswith("/"):
                main_prefix = "/" + main_prefix
            if main_prefix == "/":
                main_prefix = ""
        router._prefix_main = main_prefix
        router._prefix_main_len = len(router._prefix_main)
        router.__call__ = cls.__rcall__(router)
        cls._update_match_impl(router)
        cls._set_language_impl(router)

    @staticmethod
    def _update_match_impl(router):
        if router._impl_host_match and router._impl_scheme_match:
            router.match = router._match_w_host_and_scheme
        elif router._impl_host_match:
            router.match = router._match_w_host
        elif router._impl_scheme_match:
            router.match = router._match_w_scheme
        else:
            router.match = router._match

    @classmethod
    def _set_language_impl(cls, router):
        router.match_lang = partial(
            cls._match_w_lang if router.app.language_force_on_url else cls._match_n_lang, router
        )

    @staticmethod
    def static_versioning(router):
        return (router.app.config.static_version_urls and router.app.config.static_version) or ""

    @classmethod
    def build_route_components(cls, path):
        components = []
        params = []
        for match in cls._re_components.findall(path):
            params.append(match[1] + "{}")
        statics = cls._re_components.sub("{}", path).split("{}")
        if not params:
            components = statics
        else:
            components.append(statics[0])
            for idx, _ in enumerate(params):
                components.append(params[idx] + statics[idx + 1])
        return components

    @staticmethod
    def _match_w_lang(router, wrapper, path):
        path, lang = RouterMixin._split_lang(router, path)
        router.current.language = wrapper.language = lang
        return path

    @staticmethod
    def _match_n_lang(router, wrapper, path):
        wrapper.language = None
        return path

    @staticmethod
    def _split_lang(router, path):
        default = router.app.language_default
        if len(path) <= 1:
            return path, default
        clean_path = path.lstrip("/")
        next_sep = clean_path[2:3]
        if not next_sep or next_sep == "/":
            lang, new_path = clean_path[:2], (clean_path[2:] or "/")
            if lang != default and lang in router.app._languages_set:
                return new_path, lang
        return path, default

    @classmethod
    def __rcall__(cls, router):
        def inner(*args, **kwargs):
            if not cls._routing_started:
                cls._routing_started = True
                router.app.send_signal(cls._routing_signal)
            return RoutingCtx(router, router._routing_rule_cls, *args, **kwargs)

        return inner

    @classmethod
    def exposing(cls):
        return cls._routing_stack[-1]


class HTTPRouter(_HTTPRouter):
    __slots__ = [
        "__call__",
        "match",
        "match_lang",
        "_impl_host_match",
        "_impl_scheme_match",
        "_prefix_main_len",
        "_prefix_main",
        "_routes_str",
        "app",
        "current",
        "static_versioning",
        "routes_out",
        "routes",
        "pipeline",
    ]

    _mixin_cls = RouterMixin
    _routing_rule_cls = HTTPRoutingRule
    _routing_rec_builder = RouteRecReq

    _outputs = {
        "empty": EmptyResponseBuilder,
        "auto": AutoResponseBuilder,
        "bytes": BytesResponseBuilder,
        "str": StringResponseBuilder,
        "iter": IterResponseBuilder,
        "aiter": AsyncIterResponseBuilder,
        "http": MetaResponseBuilder,
    }

    def __init__(self, *args, **kwargs):
        self._mixin_cls._init_router_(self, *args, **kwargs)
        self.pipeline = []

    def add_route_str(self, route):
        self._routes_str[route.name] = "%s %s://%s%s%s -> %s" % (
            "|".join(route.methods),
            "|".join(route.schemes),
            route.hostname or "<any>",
            self._prefix_main,
            route.path,
            route.name,
        )

    def add_route(self, route):
        if route.hostname:
            self._impl_host_match = True
        if len(route.schemes) % 2:
            self._impl_scheme_match = True
            _scheme = "secure" if route.schemes[0] == "https" else "plain"
        else:
            _scheme = None
        self._mixin_cls._update_match_impl(self)
        for method in route.methods:
            if route.is_static:
                self.add_static_route(
                    self._routing_rec_builder(
                        name=route.name,
                        dispatch=route.dispatchers[method].dispatch,
                        flow_stream=route.pipeline_flow_stream,
                    ),
                    route.path,
                    method,
                    route.hostname,
                    _scheme,
                )
            else:
                self.add_re_route(
                    self._routing_rec_builder(
                        name=route.name,
                        dispatch=route.dispatchers[method].dispatch,
                        flow_stream=route.pipeline_flow_stream,
                    ),
                    route.build_regex(route.path),
                    route._argtypes,
                    method,
                    route.hostname,
                    _scheme,
                )
        self.routes_out[route.name] = {
            "host": route.hostname,
            "path": self._mixin_cls.build_route_components(route.path),
        }
        self.add_route_str(route)

    def _match(self, request):
        path = self.match_lang(request, request.path.rstrip("/") or request.path)
        return self.match_route_direct(request.method, path)

    def _match_w_scheme(self, request):
        path = self.match_lang(request, request.path.rstrip("/") or request.path)
        return self.match_route_scheme(request.scheme, request.method, path)

    def _match_w_host(self, request):
        path = self.match_lang(request, request.path.rstrip("/") or request.path)
        return self.match_route_host(request.host, request.method, path)

    def _match_w_host_and_scheme(self, request):
        path = self.match_lang(request, request.path.rstrip("/") or request.path)
        return self.match_route_all(request.host, request.scheme, request.method, path)

    async def dispatch(self, request, response):
        match, reqargs = self.match(request)
        if not match:
            raise HTTPBytesResponse(404, body=b"Resource not found")
        request.name = match.name
        response._bind_flow(match.flow_stream)
        return await match.dispatch(reqargs, response)


class WebsocketRouter(_WSRouter):
    __slots__ = [
        "__call__",
        "match",
        "match_lang",
        "_impl_host_match",
        "_impl_scheme_match",
        "_prefix_main_len",
        "_prefix_main",
        "_routes_str",
        "app",
        "current",
        "static_versioning",
        "routes_out",
        "routes",
        "pipeline",
    ]

    _mixin_cls = RouterMixin
    _routing_rule_cls = WebsocketRoutingRule
    _routing_rec_builder = RouteRecWS

    def __init__(self, *args, **kwargs):
        self._mixin_cls._init_router_(self, *args, **kwargs)
        self.pipeline = []

    def add_route_str(self, route):
        self._routes_str[route.name] = "%s://%s%s%s -> %s" % (
            "|".join([{"http": "ws", "https": "ws"}[scheme] for scheme in route.schemes]),
            route.hostname or "<any>",
            self._prefix_main,
            route.path,
            route.name,
        )

    def add_route(self, route):
        if route.hostname:
            self._impl_host_match = True
        if len(route.schemes) % 2:
            self._impl_scheme_match = True
            _scheme = "secure" if route.schemes[0] == "https" else "plain"
        else:
            _scheme = None
        self._mixin_cls._update_match_impl(self)
        if route.is_static:
            self.add_static_route(
                self._routing_rec_builder(
                    name=route.name,
                    dispatch=route.dispatcher.dispatch,
                    flow_recv=route.pipeline_flow_receive,
                    flow_send=route.pipeline_flow_send,
                ),
                route.path,
                route.hostname,
                _scheme,
            )
        else:
            self.add_re_route(
                self._routing_rec_builder(
                    name=route.name,
                    dispatch=route.dispatcher.dispatch,
                    flow_recv=route.pipeline_flow_receive,
                    flow_send=route.pipeline_flow_send,
                ),
                route.build_regex(route.path),
                route._argtypes,
                route.hostname,
                _scheme,
            )
        self.routes_out[route.name] = {
            "host": route.hostname,
            "path": self._mixin_cls.build_route_components(route.path),
        }
        self.add_route_str(route)

    def _match(self, websocket):
        path = self.match_lang(websocket, websocket.path.rstrip("/") or websocket.path)
        return self.match_route_direct(path)

    def _match_w_scheme(self, websocket):
        path = self.match_lang(websocket, websocket.path.rstrip("/") or websocket.path)
        return self.match_route_scheme(websocket.scheme, path)

    def _match_w_host(self, websocket):
        path = self.match_lang(websocket, websocket.path.rstrip("/") or websocket.path)
        return self.match_route_host(websocket.host, path)

    def _match_w_host_and_scheme(self, websocket):
        path = self.match_lang(websocket, websocket.path.rstrip("/") or websocket.path)
        return self.match_route_all(websocket.host, websocket.scheme, path)

    async def dispatch(self, websocket):
        match, reqargs = self.match(websocket)
        if not match:
            raise HTTPBytesResponse(404, body=b"Resource not found")
        websocket.name = match.name
        websocket._bind_flow(match.flow_recv, match.flow_send)
        await match.dispatch(reqargs)


class RoutingCtx:
    __slots__ = ["router", "rule"]

    def __init__(self, router, rule_cls: Type[RoutingRule], *args, **kwargs):
        self.router = router
        self.rule = rule_cls(self.router, *args, **kwargs)
        self.router._mixin_cls._routing_stack.append(self.rule)

    def __call__(self, f: Callable[..., Any]) -> Callable[..., Any]:
        self.router.app.send_signal(Signals.before_route, route=self.rule, f=f)
        rv = self.rule(f)
        self.router.app.send_signal(Signals.after_route, route=self.rule)
        self.router._mixin_cls._routing_stack.pop()
        return rv


class RoutingCtxGroup:
    __slots__ = ["ctxs"]

    def __init__(self, ctxs: List[RoutingCtx]):
        self.ctxs = ctxs

    def __call__(self, f: Callable[..., Any]) -> Callable[..., Any]:
        rv = f
        for ctx in self.ctxs:
            rv = ctx(f)
        return rv



def to_bytes(obj, charset="utf8", errors="strict"):
    if obj is None:
        return None
    if isinstance(obj, (bytes, bytearray, memoryview)):
        return bytes(obj)
    if isinstance(obj, str):
        return obj.encode(charset, errors)
    raise TypeError("Expected bytes")


def to_unicode(obj, charset="utf8", errors="strict"):
    if obj is None:
        return None
    if not isinstance(obj, bytes):
        return str(obj)
    return obj.decode(charset, errors)



from __future__ import annotations

import inspect
from functools import wraps

from .pipe import Pipe
from .wrappers import (
    _wrap_flow_request_basic,
    _wrap_flow_request_complete,
    _wrap_flow_request_failure,
    _wrap_flow_request_success,
    _wrap_flow_ws_basic,
    _wrap_flow_ws_complete,
    _wrap_flow_ws_failure,
    _wrap_flow_ws_success,
)


class Pipeline:
    __slots__ = ["_method_open", "_method_close", "pipes"]
    _type_suffix = ""

    def __init__(self, pipes=[]):
        self._method_open = f"open_{self._type_suffix}"
        self._method_close = f"close_{self._type_suffix}"
        self.pipes = pipes

    @staticmethod
    def _awaitable_wrap(f):
        @wraps(f)
        async def awaitable(*args, **kwargs):
            return f(*args, **kwargs)

        return awaitable

    def __call__(self, f):
        raise NotImplementedError

    def _flow_open(self):
        rv = []
        for pipe in self.pipes:
            if pipe._pipeline_all_methods_.issuperset({"open", self._method_open}):
                raise RuntimeError(
                    f"{pipe.__class__.__name__} pipe has double open methods."
                    f" Use `open` or `{self._method_open}`, not both."
                )
            if "open" in pipe._pipeline_all_methods_:
                rv.append(pipe.open)
            if self._method_open in pipe._pipeline_all_methods_:
                rv.append(getattr(pipe, self._method_open))
        return rv

    def _flow_close(self):
        rv = []
        for pipe in reversed(self.pipes):
            if pipe._pipeline_all_methods_.issuperset({"close", self._method_close}):
                raise RuntimeError(
                    f"{pipe.__class__.__name__} pipe has double close methods."
                    f" Use `close` or `{self._method_close}`, not both."
                )
            if "close" in pipe._pipeline_all_methods_:
                rv.append(pipe.close)
            if self._method_close in pipe._pipeline_all_methods_:
                rv.append(getattr(pipe, self._method_close))
        return rv

    def _flow_stream(self):
        rv = []
        for pipe in self.pipes:
            if "on_stream" not in pipe._pipeline_all_methods_:
                continue
            rv.append(pipe.on_stream)
        return rv


class RequestPipeline(Pipeline):
    __slots__ = []
    _type_suffix = "request"

    def _get_proper_wrapper(self, pipe):
        if pipe._pipeline_all_methods_.issuperset({"on_pipe_success", "on_pipe_failure"}):
            rv = _wrap_flow_request_complete
        elif "on_pipe_success" in pipe._pipeline_all_methods_:
            rv = _wrap_flow_request_success
        elif "on_pipe_failure" in pipe._pipeline_all_methods_:
            rv = _wrap_flow_request_failure
        else:
            rv = _wrap_flow_request_basic
        return rv

    def __call__(self, f):
        if not any((inspect.iscoroutinefunction(f), inspect.isasyncgenfunction(f))):
            f = self._awaitable_wrap(f)
        for pipe in reversed(self.pipes):
            if not isinstance(pipe, Pipe):
                continue
            if not pipe._is_flow_request_responsible:
                continue
            wrapper = self._get_proper_wrapper(pipe)
            pipe_method = pipe.pipe_request if "pipe_request" in pipe._pipeline_all_methods_ else pipe.pipe
            f = wrapper(pipe_method, pipe.on_pipe_success, pipe.on_pipe_failure, f)
        return f

    def _output_type(self):
        rv = None
        for pipe in reversed(self.pipes):
            if not pipe._is_flow_request_responsible or pipe.output is None:
                continue
            rv = pipe.output
        return rv


class WebsocketPipeline(Pipeline):
    __slots__ = []
    _type_suffix = "ws"

    def _get_proper_wrapper(self, pipe):
        if pipe._pipeline_all_methods_.issuperset({"on_pipe_success", "on_pipe_failure"}):
            rv = _wrap_flow_ws_complete
        elif "on_pipe_success" in pipe._pipeline_all_methods_:
            rv = _wrap_flow_ws_success
        elif "on_pipe_failure" in pipe._pipeline_all_methods_:
            rv = _wrap_flow_ws_failure
        else:
            rv = _wrap_flow_ws_basic
        return rv

    def __call__(self, f):
        if not inspect.iscoroutinefunction(f):
            f = self._awaitable_wrap(f)
        for pipe in reversed(self.pipes):
            if not isinstance(pipe, Pipe):
                continue
            if not pipe._is_flow_ws_responsible:
                continue
            wrapper = self._get_proper_wrapper(pipe)
            pipe_method = pipe.pipe_ws if "pipe_ws" in pipe._pipeline_all_methods_ else pipe.pipe
            f = wrapper(pipe_method, pipe.on_pipe_success, pipe.on_pipe_failure, f)
        return f

    def _flow_receive(self):
        rv = []
        for pipe in self.pipes:
            if "on_receive" not in pipe._pipeline_all_methods_:
                continue
            rv.append(pipe.on_receive)
        return rv

    def _flow_send(self):
        rv = []
        for pipe in reversed(self.pipes):
            if "on_send" not in pipe._pipeline_all_methods_:
                continue
            rv.append(pipe.on_send)
        return rv



from __future__ import annotations

import asyncio
import os
import re
from typing import Awaitable, Callable, Optional, Tuple

from ...ctx import RequestContext, WSContext
from ...http.response import HTTPFileResponse, HTTPResponse, HTTPStringResponse
from ...utils import cachedprop
from .helpers import WSTransport, noop_response
from .wrappers import Request, Response, Websocket


REGEX_STATIC = re.compile(r"^/static/(?P<m>__[\w\-\.]+__/)?(?P<v>_\d+\.\d+\.\d+/)?(?P<f>.*?)$")
REGEX_STATIC_LANG = re.compile(r"^/(?P<l>\w{2}/)?static/(?P<m>__[\w\-\.]__+/)?(?P<v>_\d+\.\d+\.\d+/)?(?P<f>.*?)$")


class Handler:
    __slots__ = ["app", "current"]

    def __init__(self, app, current):
        self.app = app
        self.current = current


class RequestHandler(Handler):
    __slots__ = ["router"]

    def __init__(self, app, current):
        super().__init__(app, current)
        self._bind_router()
        self._configure_methods()

    def _bind_router(self):
        raise NotImplementedError

    def _configure_methods(self):
        raise NotImplementedError


class HTTPHandler(RequestHandler):
    __slots__ = ["pre_handler", "static_handler", "static_matcher", "__dict__"]
    wapper_cls = Request
    response_cls = Response

    def _bind_router(self):
        self.router = self.app._router_http

    def _configure_methods(self):
        self.static_matcher = (
            self._static_lang_matcher if self.app.language_force_on_url else self._static_nolang_matcher
        )
        self.static_handler = self._static_handler if self.app.config.handle_static else self.dynamic_handler
        self.pre_handler = self._prefix_handler if self.router._prefix_main else self.static_handler

    async def __call__(self, scope, protocol):
        http = await self.pre_handler(scope, protocol, scope.path)
        if coro := http.rsgi(protocol):
            if self.app.config.response_timeout is None:
                await coro
                return
            try:
                await asyncio.wait_for(coro, self.app.config.response_timeout)
            except asyncio.TimeoutError:
                self.app.log.warn(f"Timeout sending response: ({scope.path})")

    @cachedprop
    def error_handler(self) -> Callable[[], Awaitable[str]]:
        return self.exception_handler

    @cachedprop
    def exception_handler(self) -> Callable[[], Awaitable[str]]:
        return self.app.error_handlers.get(500, self._exception_handler)

    @staticmethod
    async def _http_response(code: int) -> HTTPResponse:
        return HTTPResponse(code)

    def _prefix_handler(self, scope, protocol, path: str) -> Awaitable[HTTPResponse]:
        if not path.startswith(self.router._prefix_main):
            return self._http_response(404)
        path = path[self.router._prefix_main_len :] or "/"
        return self.static_handler(scope, protocol, path)

    def _static_lang_matcher(self, path: str) -> Tuple[Optional[str], Optional[str]]:
        match = REGEX_STATIC_LANG.match(path)
        if match:
            lang, mname, version, file_name = match.group("l", "m", "v", "f")
            if mname:
                mod = self.app._modules.get(mname)
                spath = mod._static_path if mod else self.app.static_path
            else:
                spath = self.app.static_path
            static_file = os.path.join(spath, file_name)
            if lang:
                lang_file = os.path.join(spath, lang, file_name)
                if os.path.exists(lang_file):
                    static_file = lang_file
            return static_file, version
        return None, None

    def _static_nolang_matcher(self, path: str) -> Tuple[Optional[str], Optional[str]]:
        if path.startswith("/static/"):
            mname, version, file_name = REGEX_STATIC.match(path).group("m", "v", "f")
            if mname:
                mod = self.app._modules.get(mname[2:-3])
                static_file = os.path.join(mod._static_path, file_name) if mod else None
            elif file_name:
                static_file = os.path.join(self.app.static_path, file_name)
            else:
                static_file = None
            return static_file, version
        return None, None

    async def _static_response(self, file_path: str) -> HTTPFileResponse:
        return HTTPFileResponse(file_path)

    def _static_handler(self, scope, protocol, path: str) -> Awaitable[HTTPResponse]:
        static_file, _ = self.static_matcher(path)
        if static_file:
            return self._static_response(static_file)
        return self.dynamic_handler(scope, protocol, path)

    async def dynamic_handler(self, scope, protocol, path: str) -> HTTPResponse:
        request = self.__class__.wapper_cls(
            scope,
            path,
            protocol,
            max_content_length=self.app.config.request_max_content_length,
            max_multipart_size=self.app.config.request_multipart_max_size,
            body_timeout=self.app.config.request_body_timeout,
        )
        response = self.response_cls(protocol)
        ctx = RequestContext(self.app, request, response)
        ctx_token = self.current._init_(ctx)
        try:
            http = await self.router.dispatch(request, response)
        except HTTPResponse as http_exception:
            http = http_exception
            #: render error with handlers if in app
            error_handler = self.app.error_handlers.get(http.status_code)
            if error_handler:
                http = HTTPStringResponse(
                    http.status_code, await error_handler(), headers=response.headers, cookies=response.cookies
                )
        except asyncio.CancelledError:
            http = noop_response
        except Exception:
            self.app.log.exception("Application exception:")
            http = HTTPStringResponse(500, await self.error_handler(), headers=response.headers)
        finally:
            self.current._close_(ctx_token)
        return http

    async def _exception_handler(self) -> str:
        self.current.response.headers._data["content-type"] = "text/plain"
        return "Internal error"


class WSHandler(RequestHandler):
    __slots__ = ["pre_handler", "__dict__"]
    wrapper_cls = Websocket

    def _bind_router(self):
        self.router = self.app._router_ws

    def _configure_methods(self):
        self.pre_handler = self._prefix_handler if self.router._prefix_main else self.dynamic_handler

    async def __call__(self, scope, protocol):
        transport = WSTransport(protocol)
        task_transport = asyncio.create_task(self.handle_transport(transport))
        task_request = asyncio.create_task(self.handle_request(scope, transport))
        _, pending = await asyncio.wait([task_request, task_transport], return_when=asyncio.FIRST_COMPLETED)
        for task in pending:
            task.cancel()
        self._close_connection(transport)

    async def handle_transport(self, transport: WSTransport):
        await transport.accepted.wait()
        try:
            while True:
                msg = await transport.transport.receive()
                # if msg.kind == WebsocketMessageType.close:
                if msg.kind == 0:
                    transport.interrupted = True
                    break
                await transport.input.put(msg)
        # except ProtocolClosed:
        except RuntimeError:
            transport.interrupted = True

    def handle_request(self, scope, transport: WSTransport):
        return self.pre_handler(scope, transport, scope.path)

    async def _empty_awaitable(self):
        return

    def _prefix_handler(self, scope, transport: WSTransport, path: str) -> Awaitable[None]:
        if not path.startswith(self.router._prefix_main):
            transport.status = 404
            return self._empty_awaitable()
        path = path[self.router._prefix_main_len :] or "/"
        return self.dynamic_handler(scope, transport, path)

    async def dynamic_handler(self, scope, transport: WSTransport, path: str):
        ctx = WSContext(self.app, self.__class__.wrapper_cls(scope, path, transport))
        ctx_token = self.current._init_(ctx)
        try:
            await self.router.dispatch(ctx.websocket)
        except HTTPResponse as http:
            transport.status = http.status_code
        except asyncio.CancelledError:
            if not transport.interrupted:
                self.app.log.exception("Application exception:")
        except Exception:
            transport.status = 500
            self.app.log.exception("Application exception:")
        finally:
            self.current._close_(ctx_token)

    def _close_connection(self, transport: WSTransport):
        transport.protocol.close(transport.status)



from contextlib import contextmanager

import pytest

from emmett_core.app import App as _App
from emmett_core.ctx import Current, RequestContext, WSContext
from emmett_core.datastructures import sdict
from emmett_core.protocols.rsgi.test_client.scope import ScopeBuilder
from emmett_core.protocols.rsgi.wrappers import Request, Websocket
from emmett_core.routing.router import HTTPRouter, WebsocketRouter


class App(_App):
    def _init_routers(self, url_prefix):
        pass

    def _init_handlers(self):
        pass

    def _register_with_ctx(self):
        pass

    def _init_with_test_env(self, current):
        self._router_http = HTTPRouter(self, current)
        self._router_ws = WebsocketRouter(self, current)
        current.app = self


class FakeWSTransport(sdict):
    async def send_str(self, data):
        self.current._send_storage.append(data)

    async def send_bytes(self, data):
        self.current._send_storage.append(data.decode("utf8"))


class FakeWSProtocol(sdict):
    async def accept(self):
        pass

    async def init(self):
        self.transport = FakeWSTransport(current=self.current)
        return self.transport

    async def receive(self):
        return sdict(data="{}")


@pytest.fixture(scope="function")
def current():
    return Current()


@pytest.fixture(scope="function")
def app(current):
    rv = App(__name__)
    rv._init_with_test_env(current)
    return rv


@pytest.fixture(scope="function")
def http_ctx_builder(current):
    @contextmanager
    def ctx_builder(path):
        scope = ScopeBuilder(path=path).get_data()[0]
        token = current._init_(
            RequestContext(current.app, Request(scope, scope.path, None), sdict(_bind_flow=lambda v: None))
        )
        yield sdict(ctx=current, wrapper=current.request)
        current._close_(token)

    return ctx_builder


@pytest.fixture(scope="function")
def ws_ctx_builder(current):
    @contextmanager
    def ctx_builder(path):
        scope = ScopeBuilder(path=path).get_data()[0]
        scope.proto = "ws"
        token = current._init_(WSContext(current.app, Websocket(scope, scope.path, FakeWSProtocol(current=current))))
        yield sdict(ctx=current, wrapper=current.websocket)
        current._close_(token)

    return ctx_builder



from __future__ import annotations

import asyncio
import copy
from io import BytesIO

from ....ctx import Current, RequestContext
from ....http.response import HTTPResponse, HTTPStringResponse
from ....parsers import Parsers
from ....utils import cachedprop
from ..handlers import HTTPHandler
from ..wrappers import Request, Response
from .helpers import Headers, TestCookieJar
from .scope import ScopeBuilder
from .urls import get_host, url_parse, url_unparse


class ClientContextResponse(Response):
    def __init__(self, original_response: Response):
        super().__init__(original_response._proto)
        self.status = original_response.status
        self.headers._data.update(original_response.headers._data)
        self.cookies.update(original_response.cookies.copy())
        self.__dict__.update(original_response.__dict__)


class ClientContext:
    _response_wrap_cls = ClientContextResponse

    def __init__(self, ctx):
        self.request = Request(ctx.request._scope, ctx.request.path, None)
        self.response = self.__class__._response_wrap_cls(ctx.response)
        self.session = copy.deepcopy(ctx.session)

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_value, tb):
        pass


class ClientHTTPHandlerMixin:
    _client_ctx_cls = ClientContext

    async def dynamic_handler(self, scope, protocol, path):
        request = self.__class__.wapper_cls(
            scope,
            path,
            protocol,
            max_content_length=self.app.config.request_max_content_length,
            max_multipart_size=self.app.config.request_multipart_max_size,
            body_timeout=self.app.config.request_body_timeout,
        )
        response = self.response_cls(protocol)
        ctx = RequestContext(self.app, request, response)
        ctx_token = self.current._init_(ctx)
        try:
            http = await self.router.dispatch(request, response)
        except HTTPResponse as http_exception:
            http = http_exception
            #: render error with handlers if in app
            error_handler = self.app.error_handlers.get(http.status_code)
            if error_handler:
                http = HTTPStringResponse(
                    http.status_code, await error_handler(), headers=response.headers, cookies=response.cookies
                )
        except Exception:
            self.app.log.exception("Application exception:")
            http = HTTPStringResponse(500, await self.error_handler(), headers=response.headers)
        finally:
            scope._ctx = self.__class__._client_ctx_cls(ctx)
            self.current._close_(ctx_token)
        return http


class ClientHTTPHandler(ClientHTTPHandlerMixin, HTTPHandler): ...


class ClientResponse:
    def __init__(self, ctx, raw, status, headers):
        self.context = ctx
        self.raw = raw
        self.status = status
        self.headers = headers
        self._close = lambda: None

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_value, tb):
        pass

    @cachedprop
    def data(self):
        if isinstance(self.raw, bytes):
            return self.raw.decode("utf8")
        return self.raw

    def json(self):
        return Parsers.get_for("json")(self.data)


class EmmettTestClient:
    _current: Current
    _handler_cls = ClientHTTPHandler

    def __init__(self, application, response_wrapper=ClientResponse, use_cookies=True, allow_subdomain_redirects=False):
        self.application = application
        self.response_wrapper = response_wrapper
        if use_cookies:
            self.cookie_jar = TestCookieJar()
        else:
            self.cookie_jar = None
        self.allow_subdomain_redirects = allow_subdomain_redirects

    def run_rsgi_app(self, scope, body):
        if self.cookie_jar is not None:
            self.cookie_jar.inject_rsgi(scope)
        rv = run_rsgi_app(
            self.__class__._current, self.application, scope, body, _handler_cls=self.__class__._handler_cls
        )
        if self.cookie_jar is not None:
            self.cookie_jar.extract_rsgi(scope, Headers(rv["headers"]))
        return rv

    def resolve_redirect(self, response, new_loc, scope, headers):
        scheme, netloc, script_root, qs, _anchor = url_parse(new_loc)
        base_url = url_unparse((scheme, netloc, "", "", "")).rstrip("/") + "/"

        cur_name = netloc.split(":", 1)[0].split(".")
        real_name = get_host(scope, headers).rsplit(":", 1)[0].split(".")

        if len(cur_name) == 1 and not cur_name[0]:
            allowed = True
        else:
            if self.allow_subdomain_redirects:
                allowed = cur_name[-len(real_name) :] == real_name
            else:
                allowed = cur_name == real_name

        if not allowed:
            raise RuntimeError("%r does not support redirect to external targets" % self.__class__)

        status_code = response["status"]
        if status_code == 307:
            method = scope.method
        else:
            method = "GET"

        # For redirect handling we temporarily disable the response
        # wrapper.  This is not threadsafe but not a real concern
        # since the test client must not be shared anyways.
        old_response_wrapper = self.response_wrapper
        self.response_wrapper = None
        try:
            return self.open(path=script_root, base_url=base_url, query_string=qs, method=method, as_tuple=True)
        finally:
            self.response_wrapper = old_response_wrapper

    def open(self, *args, **kwargs):
        as_tuple = kwargs.pop("as_tuple", False)
        follow_redirects = kwargs.pop("follow_redirects", False)
        scope, body = None, b""
        if not kwargs and len(args) == 1:
            if isinstance(args[0], ScopeBuilder):
                scope, body = args[0].get_data()
        if scope is None:
            builder = ScopeBuilder(*args, **kwargs)
            try:
                scope, body = builder.get_data()
            finally:
                builder.close()

        response = self.run_rsgi_app(scope, body)

        # handle redirects
        redirect_chain = []
        while 1:
            status_code = response["status"]
            if status_code not in (301, 302, 303, 305, 307) or not follow_redirects:
                break
            headers = Headers(response["headers"])
            new_location = headers["location"]
            if new_location.startswith("/"):
                new_location = scope.scheme + "://" + scope.server + new_location
            new_redirect_entry = (new_location, status_code)
            if new_redirect_entry in redirect_chain:
                raise Exception("loop detected")
            redirect_chain.append(new_redirect_entry)
            scope, response = self.resolve_redirect(response, new_location, scope, headers)

        if self.response_wrapper is not None:
            response = self.response_wrapper(
                scope._ctx, response["body"], response["status"], Headers(response["headers"])
            )
        if as_tuple:
            return scope, response
        return response

    def get(self, *args, **kw):
        kw["method"] = "GET"
        return self.open(*args, **kw)

    def patch(self, *args, **kw):
        kw["method"] = "PATCH"
        return self.open(*args, **kw)

    def post(self, *args, **kw):
        kw["method"] = "POST"
        return self.open(*args, **kw)

    def head(self, *args, **kw):
        kw["method"] = "HEAD"
        return self.open(*args, **kw)

    def put(self, *args, **kw):
        kw["method"] = "PUT"
        return self.open(*args, **kw)

    def delete(self, *args, **kw):
        kw["method"] = "DELETE"
        return self.open(*args, **kw)

    def options(self, *args, **kw):
        kw["method"] = "OPTIONS"
        return self.open(*args, **kw)

    def trace(self, *args, **kw):
        kw["method"] = "TRACE"
        return self.open(*args, **kw)

    def __repr__(self):
        return "<%s %r>" % (self.__class__.__name__, self.application)


class ClientHTTPStreamTransport:
    def __init__(self):
        self._data = BytesIO()

    async def send_bytes(self, data: bytes):
        self._data.write(data)

    async def send_str(self, data: str):
        self._data.write(data.encode("utf8"))


class ClientHTTPProtocol:
    def __init__(self, body):
        self.input = body
        self.response_status = 500
        self.response_headers = []
        self.response_body = b""
        self.response_body_stream = None
        self.consumed_input = False
        self.input_step = 0

    async def __call__(self):
        self.consumed_input = True
        return self.input

    def __aiter__(self):
        return self

    async def __anext__(self):
        if self.consumed_input:
            raise StopAsyncIteration
        input_range = (4096 * self.input_step, 4096 * (self.input_step + 1))
        ret = self.input[input_range[0] : input_range[1]]
        if not ret:
            self.consumed_input = True
            raise StopAsyncIteration
        self.input_step += 1
        return ret

    def response_empty(self, status, headers):
        self.response_status = status
        self.response_headers = headers

    def response_str(self, status, headers, body):
        self.response_status = status
        self.response_headers = headers
        self.response_body = body.encode("utf8")

    def response_bytes(self, status, headers, body):
        self.response_status = status
        self.response_headers = headers
        self.response_body = body

    def response_file(self, status, headers, file):
        self.response_status = status
        self.response_headers = headers
        with open(file, "rb") as f:
            self.response_body = f.read()

    def response_stream(self, status, headers):
        self.response_status = status
        self.response_headers = headers
        self.response_body_stream = ClientHTTPStreamTransport()
        return self.response_body_stream


def run_rsgi_app(current, app, scope, body=b"", _handler_cls=ClientHTTPHandler):
    try:
        loop = asyncio.get_event_loop()
    except RuntimeError:
        loop = asyncio.new_event_loop()
        asyncio.set_event_loop(loop)

    raw = {}
    proto = ClientHTTPProtocol(body)
    handler = _handler_cls(app, current)
    loop.run_until_complete(handler(scope, proto))
    raw["version"] = 11
    raw["status"] = proto.response_status
    raw["headers"] = proto.response_headers
    if proto.response_body_stream:
        proto.response_body_stream._data.seek(0)
        raw["body"] = proto.response_body_stream._data.read()
    else:
        raw["body"] = proto.response_body

    return raw



from enum import Enum

from .. import _emmett_core


AES_BLOCK_SIZE = 128 // 8


class AESModes(Enum):
    CFB8 = {
        128 // 8: (_emmett_core.aes128_cfb8_encrypt, _emmett_core.aes128_cfb8_decrypt),
        256 // 8: (_emmett_core.aes256_cfb8_encrypt, _emmett_core.aes256_cfb8_decrypt),
    }
    CFB128 = {
        128 // 8: (_emmett_core.aes128_cfb128_encrypt, _emmett_core.aes128_cfb128_decrypt),
        256 // 8: (_emmett_core.aes256_cfb128_encrypt, _emmett_core.aes256_cfb128_decrypt),
    }
    CTR128 = {
        128 // 8: (_emmett_core.aes128_ctr128, _emmett_core.aes128_ctr128),
        256 // 8: (_emmett_core.aes256_ctr128, _emmett_core.aes256_ctr128),
    }


def aes_encrypt(data: bytes, key: bytes, nonce: bytes, mode: AESModes = AESModes.CTR128) -> bytes:
    try:
        method, _ = mode.value[len(key)]
    except KeyError:
        raise ValueError("key must be 16 or 32 bytes long")
    assert len(nonce) == AES_BLOCK_SIZE, f"nonce must be {AES_BLOCK_SIZE} bytes long"
    return method(data, key, nonce)


def aes_decrypt(data: bytes, key: bytes, nonce: bytes, mode: AESModes = AESModes.CTR128) -> bytes:
    try:
        _, method = mode.value[len(key)]
    except KeyError:
        raise ValueError("key must be 16 or 32 bytes long")
    assert len(nonce) == AES_BLOCK_SIZE, f"nonce must be {AES_BLOCK_SIZE} bytes long"
    return method(data, key, nonce)



import pytest

from emmett_core.routing.router import HTTPRouter
from emmett_core.routing.urls import Url


def route(router, path, **kwargs):
    def wrap(f):
        return router(paths=path, **kwargs)(f)

    return wrap


def cfg_router(router):
    @route(router, "/test_complex/<int:a>/<float:b>/<date:c>/<alpha:d>/<str:e>/<any:f>")
    def test_route_complex(a, b, c, d, e, f):
        return

    @route(router, "/test_optionals/<int:a>/foo(/<str:b>)?(.<str:c>)?")
    def test_route_optionals(a, b, c):
        return

    return router


@pytest.fixture(scope="function")
def url(current):
    return Url(current)


@pytest.fixture(scope="function")
def app_static_version(app):
    app.config.static_version_urls = True
    app.config.static_version = "1.0.0"
    return app


@pytest.fixture(scope="function")
def http_router_prefix(app, current):
    rv = HTTPRouter(app, current, url_prefix="/test_prefix")
    app._router_http = rv
    return rv


@pytest.fixture(scope="function")
def router(request, http_router, ws_router):
    return {"http": cfg_router(http_router), "ws": cfg_router(ws_router)}[request.param]


def test_static(url, http_router):
    link = url("static", "file")
    assert link == "/static/file"


def test_static_version(url, app_static_version, http_router):
    link = url("static", "js/foo.js", language="it")
    assert link == "/it/static/_1.0.0/js/foo.js"


def test_prefix(url, http_router_prefix):
    link = url("static", "js/foo.js", language="it")
    assert link == "/test_prefix/it/static/js/foo.js"


@pytest.mark.parametrize(("router"), ["http"], indirect=True)
@pytest.mark.parametrize(
    ("parts", "expected"),
    [
        (
            ("test_urls.test_route_complex", [1, 1.2, "2000-12-01", "foo", "foo1", "bar", "baz"], {}, {}),
            "/test_complex/1/1.2/2000-12-01/foo/foo1/bar/baz",
        ),
        (
            ("test_urls.test_route_complex", [1, 1.2, "2000-12-01", "foo", "foo1", "bar", "baz"], {"foo": "bar"}, {}),
            "/test_complex/1/1.2/2000-12-01/foo/foo1/bar/baz?foo=bar",
        ),
        (
            ("test_urls.test_route_optionals", [1], {}, {}),
            "/test_optionals/1/foo",
        ),
        (
            ("test_urls.test_route_optionals", [1, "bar"], {}, {}),
            "/test_optionals/1/foo/bar",
        ),
        (
            ("test_urls.test_route_optionals", [1, "bar", "baz"], {}, {}),
            "/test_optionals/1/foo/bar.baz",
        ),
        (
            ("test_urls.test_route_optionals", [1, "bar"], {"foo": "bar"}, {}),
            "/test_optionals/1/foo/bar?foo=bar",
        ),
        (
            ("test_urls.test_route_optionals", [1], {}, {"scheme": True}),
            "http://localhost/test_optionals/1/foo",
        ),
        (
            ("test_urls.test_route_optionals", [1], {}, {"scheme": True, "host": "test1.tld"}),
            "http://test1.tld/test_optionals/1/foo",
        ),
        (
            ("test_urls.test_route_optionals", [1], {}, {"scheme": "https"}),
            "https://localhost/test_optionals/1/foo",
        ),
    ],
)
def test_req(http_ctx_builder, url, router, parts, expected):
    route_name, args, params, extras = parts
    with http_ctx_builder("/test"):
        assert url(route_name, args, params, **extras) == expected


@pytest.mark.parametrize(("router"), ["ws"], indirect=True)
@pytest.mark.parametrize(
    ("parts", "expected"),
    [
        (
            ("test_urls.test_route_complex", [1, 1.2, "2000-12-01", "foo", "foo1", "bar", "baz"], {}, {}),
            "/test_complex/1/1.2/2000-12-01/foo/foo1/bar/baz",
        ),
        (
            ("test_urls.test_route_complex", [1, 1.2, "2000-12-01", "foo", "foo1", "bar", "baz"], {"foo": "bar"}, {}),
            "/test_complex/1/1.2/2000-12-01/foo/foo1/bar/baz?foo=bar",
        ),
        (
            ("test_urls.test_route_optionals", [1], {}, {}),
            "/test_optionals/1/foo",
        ),
        (
            ("test_urls.test_route_optionals", [1, "bar"], {}, {}),
            "/test_optionals/1/foo/bar",
        ),
        (
            ("test_urls.test_route_optionals", [1, "bar", "baz"], {}, {}),
            "/test_optionals/1/foo/bar.baz",
        ),
        (
            ("test_urls.test_route_optionals", [1, "bar"], {"foo": "bar"}, {}),
            "/test_optionals/1/foo/bar?foo=bar",
        ),
        (
            ("test_urls.test_route_optionals", [1], {}, {"scheme": True}),
            "ws://localhost/test_optionals/1/foo",
        ),
        (
            ("test_urls.test_route_optionals", [1], {}, {"scheme": True, "host": "test1.tld"}),
            "ws://test1.tld/test_optionals/1/foo",
        ),
        (
            ("test_urls.test_route_optionals", [1], {}, {"scheme": "https"}),
            "wss://localhost/test_optionals/1/foo",
        ),
        (
            ("test_urls.test_route_optionals", [1], {}, {"scheme": "wss"}),
            "wss://localhost/test_optionals/1/foo",
        ),
    ],
)
def test_ws(http_ctx_builder, url, router, parts, expected):
    route_name, args, params, extras = parts

    with http_ctx_builder("/test"):
        assert url.ws(route_name, args, params, **extras) == expected



from __future__ import annotations

from typing import Any, Callable, Dict, List, Union

from ..cache.handlers import CacheHandler
from ..cache.hash import CacheHashMixin
from .dispatchers import RequestDispatcher


class RouteCacheRule(CacheHashMixin):
    def __init__(
        self,
        handler: CacheHandler,
        query_params: bool = True,
        language: bool = True,
        hostname: bool = False,
        headers: List[str] = [],
        duration: Union[int, str, None] = "default",
    ):
        super().__init__()
        self.cache = handler
        self.check_headers = headers
        self.duration = duration
        self.add_strategy("kwargs", self.dict_strategy)
        self._ctx_builders = []
        if hostname:
            self.add_strategy("hostname")
            self._ctx_builders.append(("hostname", lambda route, current: route.hostname))
        if language:
            self.add_strategy("language")
            self._ctx_builders.append(("language", lambda route, current: current.language))
        if query_params:
            self.add_strategy("query_params", self.dict_strategy)
            self._ctx_builders.append(("query_params", lambda route, current: current.request.query_params))
        if headers:
            self.add_strategy("headers", self.headers_strategy)
            self._ctx_builders.append(("headers", lambda route, current: current.request.headers))

    def _build_ctx_key(self, route: Any, **ctx) -> str:  # type: ignore
        return route.name + ":" + self._build_hash(ctx)

    def _build_ctx(self, kwargs: Dict[str, Any], route: Any, current: Any) -> Dict[str, Any]:
        rv = {"kwargs": kwargs}
        for key, builder in self._ctx_builders:
            rv[key] = builder(route, current)
        return rv

    def headers_strategy(self, data: Dict[str, str]) -> List[str]:
        return [data[key] for key in self.check_headers]

    def __call__(self, f: Callable[..., Any]) -> Callable[..., Any]:
        from .router import RouterMixin

        obj = RouterMixin.exposing()
        obj.cache_rule = self
        return f


class CacheDispatcher(RequestDispatcher):
    __slots__ = ["current", "route", "cache_rule"]

    def __init__(self, route, rule, response_builder):
        super().__init__(route, rule, response_builder)
        self.current = rule.current
        self.route = route
        self.cache_rule = rule.cache_rule

    async def get_data(self, reqargs, response):
        key = self.cache_rule._build_ctx_key(
            self.route, **self.cache_rule._build_ctx(reqargs, self.route, self.current)
        )
        data = self.cache_rule.cache.get(key)
        if data is not None:
            response.headers.update(data["headers"])
            return data["content"]
        content = await self.f(**reqargs)
        if response.status == 200:
            self.cache_rule.cache.set(key, {"content": content, "headers": response.headers}, self.cache_rule.duration)
        return content

    async def dispatch(self, reqargs, response):
        content = await self.get_data(reqargs, response)
        return self.response_builder(content, response)


class CacheOpenDispatcher(CacheDispatcher):
    __slots__ = []

    async def dispatch(self, reqargs, response):
        await self._parallel_flow(self.flow_open)
        return await super().dispatch(reqargs, response)


class CacheCloseDispatcher(CacheDispatcher):
    __slots__ = []

    async def dispatch(self, reqargs, response):
        try:
            content = await self.get_data(reqargs, response)
        except Exception:
            await self._parallel_flow(self.flow_close)
            raise
        await self._parallel_flow(self.flow_close)
        return self.response_builder(content, response)


class CacheFlowDispatcher(CacheDispatcher):
    __slots__ = []

    async def dispatch(self, reqargs, response):
        await self._parallel_flow(self.flow_open)
        try:
            content = await self.get_data(reqargs, response)
        except Exception:
            await self._parallel_flow(self.flow_close)
            raise
        await self._parallel_flow(self.flow_close)
        return self.response_builder(content, response)



import re

from .cache import (
    CacheCloseDispatcher,
    CacheDispatcher,
    CacheFlowDispatcher,
    CacheOpenDispatcher,
)
from .dispatchers import (
    Dispatcher,
    RequestCloseDispatcher,
    RequestDispatcher,
    RequestFlowDispatcher,
    RequestOpenDispatcher,
    WSCloseDispatcher,
    WSFlowDispatcher,
    WSOpenDispatcher,
)


REGEX_INT = re.compile(r"<int\:(\w+)>")
REGEX_STR = re.compile(r"<str\:(\w+)>")
REGEX_ANY = re.compile(r"<any\:(\w+)>")
REGEX_ALPHA = re.compile(r"<alpha\:(\w+)>")
REGEX_DATE = re.compile(r"<date\:(\w+)>")
REGEX_FLOAT = re.compile(r"<float\:(\w+)>")


class Route:
    __slots__ = [
        "f",
        "hostname",
        "is_static",
        "name",
        "path",
        "pipeline_flow_close",
        "pipeline_flow_open",
        "schemes",
        "_argtypes",
    ]
    _re_condl = re.compile(r"\(.*\)\?")
    _re_param = re.compile(r"<(\w+)\:(\w+)>")

    def __init__(self, rule, path, idx):
        self.name = rule.name if idx == 0 else f"{rule.name}.{idx}"
        self.f = rule.f
        if not path.startswith("/"):
            path = "/" + path
        if rule.prefix:
            path = (path != "/" and rule.prefix + path) or rule.prefix
        self.path = path
        self.schemes = tuple(rule.schemes)
        self.hostname = rule.hostname
        self.pipeline_flow_open = rule.pipeline_flow_open
        self.pipeline_flow_close = rule.pipeline_flow_close
        self._init_match_type()
        self._build_argtmap()

    def _init_match_type(self):
        if self._re_condl.findall(self.path) or self._re_param.findall(self.path):
            self.is_static = False
        else:
            self.is_static = True

    def _build_argtmap(self):
        types = {"int", "float", "date"}
        tmap = {}
        for argtype in types:
            for arg in set(re.compile(r"<{}\:(\w+)>".format(argtype)).findall(self.path)):
                tmap[arg] = argtype
        self._argtypes = tmap

    @staticmethod
    def build_regex(path):
        path = REGEX_INT.sub(r"(?P<\g<1>>\\d+)", path)
        path = REGEX_STR.sub(r"(?P<\g<1>>[^/]+)", path)
        path = REGEX_ANY.sub(r"(?P<\g<1>>.*)", path)
        path = REGEX_ALPHA.sub(r"(?P<\g<1>>[^/\\W\\d_]+)", path)
        path = REGEX_DATE.sub(r"(?P<\g<1>>\\d{4}-\\d{2}-\\d{2})", path)
        path = REGEX_FLOAT.sub(r"(?P<\g<1>>\\d+\.\\d+)", path)
        return f"^{path}$"


class HTTPRoute(Route):
    __slots__ = ["methods", "pipeline_flow_stream", "dispatchers"]

    def __init__(self, rule, path, idx):
        super().__init__(rule, path, idx)
        self.methods = tuple(method.upper() for method in rule.methods)
        self.pipeline_flow_stream = rule.pipeline_flow_stream
        dispatchers = {
            "base": (RequestDispatcher, CacheDispatcher),
            "open": (RequestOpenDispatcher, CacheOpenDispatcher),
            "close": (RequestCloseDispatcher, CacheCloseDispatcher),
            "flow": (RequestFlowDispatcher, CacheFlowDispatcher),
        }
        if self.pipeline_flow_open and self.pipeline_flow_close:
            dispatcher, cdispatcher = dispatchers["flow"]
        elif self.pipeline_flow_open and not self.pipeline_flow_close:
            dispatcher, cdispatcher = dispatchers["open"]
        elif not self.pipeline_flow_open and self.pipeline_flow_close:
            dispatcher, cdispatcher = dispatchers["close"]
        else:
            dispatcher, cdispatcher = dispatchers["base"]
        self.dispatchers = {}
        for method in self.methods:
            dispatcher_cls = cdispatcher if rule.cache_rule and method in ["HEAD", "GET"] else dispatcher
            self.dispatchers[method] = dispatcher_cls(
                self, rule, rule.head_builder if method == "HEAD" else rule.response_builder
            )


class WebsocketRoute(Route):
    __slots__ = ["pipeline_flow_receive", "pipeline_flow_send", "dispatcher"]

    def __init__(self, rule, path, idx):
        super().__init__(rule, path, idx)
        self.pipeline_flow_receive = rule.pipeline_flow_receive
        self.pipeline_flow_send = rule.pipeline_flow_send
        dispatchers = {
            "base": Dispatcher,
            "open": WSOpenDispatcher,
            "close": WSCloseDispatcher,
            "flow": WSFlowDispatcher,
        }
        if self.pipeline_flow_open and self.pipeline_flow_close:
            dispatcher = dispatchers["flow"]
        elif self.pipeline_flow_open and not self.pipeline_flow_close:
            dispatcher = dispatchers["open"]
        elif not self.pipeline_flow_open and self.pipeline_flow_close:
            dispatcher = dispatchers["close"]
        else:
            dispatcher = dispatchers["base"]
        self.dispatcher = dispatcher(self)



from functools import partial
from typing import Any, Callable, Dict, Union

from ._imports import orjson, rapidjson


if orjson is not None:
    _json_impl = orjson.dumps
    _json_opts = {"option": orjson.OPT_NON_STR_KEYS | orjson.OPT_NAIVE_UTC}
    _json_type = "bytes"
elif rapidjson is not None:
    _json_impl = rapidjson.dumps
    _json_opts = {
        "datetime_mode": rapidjson.DM_ISO8601 | rapidjson.DM_NAIVE_IS_UTC,
        "number_mode": rapidjson.NM_NATIVE | rapidjson.NM_DECIMAL,
    }
    _json_type = "str"
else:
    import json as _stdjson

    _json_impl = _stdjson.dumps
    _json_opts = {}
    _json_type = "str"

_json_safe_table = {"u2028": [r"\u2028", "\\u2028"], "u2029": [r"\u2029", "\\u2029"]}


class Serializers:
    _registry_: Dict[str, Callable[[Any], Union[bytes, str]]] = {}

    @classmethod
    def register_for(cls, target):
        def wrap(f):
            cls._registry_[target] = f
            return f

        return wrap

    @classmethod
    def get_for(cls, target):
        return cls._registry_[target]


def _json_default(obj):
    if hasattr(obj, "__json__"):
        return obj.__json__()
    raise TypeError


json = partial(_json_impl, default=_json_default, **_json_opts)


def json_safe(value):
    rv = json(value)
    for val, rep in _json_safe_table.values():
        rv.replace(val, rep)
    return rv


Serializers.register_for("json")(json)



from __future__ import annotations

from enum import Enum
from functools import wraps
from typing import Any, Callable, Dict, Optional, TypeVar

from .datastructures import sdict


class Signals(str, Enum):
    __str__ = lambda v: v.value

    after_loop = "after_loop"
    after_route = "after_route"
    before_route = "before_route"
    before_routes = "before_routes"


class listen_signal:
    _inst_count_ = 0

    def __init__(self, signal: Signals):
        self.signal = signal
        self._inst_count_ = listen_signal._inst_count_
        listen_signal._inst_count_ += 1

    def __call__(self, f: Callable[..., None]) -> listen_signal:
        self.f = f
        return self


class MetaExtension(type):
    def __new__(cls, name, bases, attrs):
        new_class = type.__new__(cls, name, bases, attrs)
        declared_listeners = {}
        all_listeners = {}
        listeners = []
        for key, value in list(attrs.items()):
            if isinstance(value, listen_signal):
                listeners.append((key, value))
        listeners.sort(key=lambda x: x[1]._inst_count_)
        declared_listeners.update(listeners)
        new_class._declared_listeners_ = declared_listeners
        for base in reversed(new_class.__mro__[1:]):
            if hasattr(base, "_declared_listeners_"):
                all_listeners.update(base._declared_listeners_)
        all_listeners.update(declared_listeners)
        new_class._all_listeners_ = all_listeners
        return new_class


class Extension(metaclass=MetaExtension):
    namespace: Optional[str] = None
    default_config: Dict[str, Any] = {}

    def __init__(self, app, env: sdict, config: sdict):
        self.app = app
        self.env = env
        self.config = config
        self.__init_config()
        self.__init_listeners()

    def __init_config(self):
        for key, dval in self.default_config.items():
            self.config[key] = self.config.get(key, dval)

    def __init_listeners(self):
        self._listeners_ = []
        for _name, obj in self._all_listeners_.items():
            self._listeners_.append((str(obj.signal), _wrap_listener(self, obj.f)))

    def on_load(self):
        pass


def _wrap_listener(ext, f):
    @wraps(f)
    def wrapped(*args, **kwargs):
        return f(ext, *args, **kwargs)

    return wrapped


ExtensionType = TypeVar("ExtensionType", bound=Extension)
